%THETANOTE2
%   Computational experiment (benefits of adaptivity) for the paper
%   "On well-separated sets and fast multipole methods". Graphics in
%   THETANOTEPLOTS.

% S. Engblom 2011-06-29 (Minor revision)
% S. Engblom 2011-01-28

% NOTE: to perform this experiment the old uniform code fmm2du has to be
% properly compiled and placed on the Matlab-path.

warning('Platform-dependent make follows...');
% in order not to reach the unfair conclusion that the adaptive version
% fmm2d is always faster than the uniform code fmm2du, one must compile the
% former with #undef FMM_INLINECOMPLEX, and #undef FMM_INLINECOMPLEX_DIRECT;
% this statement is platform-dependent, see make.m
mx = mexext;
if strcmp(mx,'mexa64') || strcmp(mx,'mexmaci64')
  make('-DBLASINT=size_t','-DFMM_SHRINK');
else
  make('-DFMM_SHRINK');
end

% check the number of levels (should be equal for fairness)
N = 200000;
Ndirect1 = 10;
Ndirect2 = 10;
ceil(0.5*log2(5/8*N/Ndirect1)) % fmm2d
ceil(0.5*log2(N/Ndirect2))-1   % fmm2du

% check the number of coefficients (should also be equal)
tol1 = 2e-7;
tol2 = 1e-6;
floor(log2(4*tol1)/log2(0.5)-1)+1 % fmm2d with [theta,pot] = [0.5,1]
floor(-log2(tol2))+1              % fmm2du with pot = 1

% range of shape parameter
sigma = 1./2.^linspace(-2,5,15);

ta1 = zeros(size(sigma));
tu1 = zeros(size(sigma));
ta2 = zeros(size(sigma));
tu2 = zeros(size(sigma));

% to get accurate timings and a smooth graph
Ntrymax = 9;
Ntrymin = 5;
Tmin = 5;
m = ones(N,1);

for i = 1:numel(sigma)
  randn('state',20110202);
  % normal distribution N(0,sigma), enforced to occupy exactly the unit
  % square (by rejection)
  x = [0; 1]; % enforce the unit square
  y = [1; 0];
  while numel(x) < N || numel(y) < N
    x = [abs(sigma(i)*randn(N-numel(x),1)); x];
    y = [y; abs(sigma(i)*randn(N-numel(y),1))];
    x(x > 1) = [];
    y(y > 1) = [];
  end
  z = complex(x,y);

  % adaptive version
  j = 0;
  T = 0;
  tt = zeros(1,Ntrymax);
  while j < Ntrymin || (T < Tmin && j < Ntrymax)
    j = j+1;
    tt_ = cputime;
    p = fmm2d(z,m,'pot',1,'tol',tol1,'ndirect',Ndirect1);
    tt(j) = cputime-tt_;
    T = T+tt(j);
  end
  % form a robust mean - "trimean with discarded outliers" - to get a
  % smooth graph
  tt = sort(tt(1:j)); tt([1 end]) = [];
  % [1:j-2] = [1:a a+1:b b+1:j-2] with the first and last interval of
  % equal length
  a = floor((j-2)/3); b = j-2-a;
  ta1(i) = 0.5*mean(tt([1:a b+1:end]))+0.5*mean(tt(a+1:b));

  % uniform version
  j = 0;
  T = 0;
  tt = zeros(1,Ntrymax);
  while j < Ntrymin || (T < Tmin && j < Ntrymax)
    j = j+1;
    tt_ = cputime;
    p = fmm2du(z,m,'pot',1,'tol',tol2,'ndirect',Ndirect2);
    tt(j) = cputime-tt_;
    T = T+tt(j);
  end
  tt = sort(tt(1:j)); tt([1 end]) = [];
  a = floor((j-2)/3); b = j-2-a;
  tu1(i) = 0.5*mean(tt([1:a b+1:end]))+0.5*mean(tt(a+1:b));
  
  % as above, but the x-coordinate is now uniformly distributed so that
  % an artificial "boundary layer" results
  z = complex([rand(N-2,1); [0; 1]],imag(z));

  % adaptive version
  j = 0;
  T = 0;
  tt = zeros(1,Ntrymax);
  while j < Ntrymin || (T < Tmin && j < Ntrymax)
    j = j+1;
    tt_ = cputime;
    p = fmm2d(z,m,'pot',1,'tol',tol1,'ndirect',Ndirect1);
    tt(j) = cputime-tt_;
    T = T+tt(j);
  end
  tt = sort(tt(1:j)); tt([1 end]) = [];
  a = floor((j-2)/3); b = j-2-a;
  ta2(i) = 0.5*mean(tt([1:a b+1:end]))+0.5*mean(tt(a+1:b));

  % uniform version
  j = 0;
  T = 0;
  tt = zeros(1,Ntrymax);
  while j < Ntrymin || (T < Tmin && j < Ntrymax)
    j = j+1;
    tt_ = cputime;
    p = fmm2du(z,m,'pot',1,'tol',tol2,'ndirect',Ndirect2);
    tt(j) = cputime-tt_;
    T = T+tt(j);
  end
  tt = sort(tt(1:j)); tt([1 end]) = [];
  a = floor((j-2)/3); b = j-2-a;
  tu2(i) = 0.5*mean(tt([1:a b+1:end]))+0.5*mean(tt(a+1:b));
  i
end

if exist('./examples/thetanote2.mat','file')
  fprintf('Existing file thetanote2.mat NOT overwritten.\n');
else
  save examples/thetanote2 sigma ta1 tu1 ta2 tu2
end

make % compile back using defaults

% below what sigma is it beneficial to use adaptivity?
cutoff1 = interp1(ta1./tu1,sigma,1)
cutoff2 = interp1(ta2./tu2,sigma,1)
