function [p,pe] = fmm2d(z,m,ze,varargin)
%FMM2D Fast multipole method in 2 dimensions.
%   P = FMM2D(Z,M) evaluates the potential P due to masses M at
%   Cartesian coordinates Z = X+iY. It is understood that the size of
%   Z matches that of M.
%
%   PE = FMM2D(Z,M,ZE) evaluates the potential at the points ZE
%   instead.
%
%   [P,PE] = FMM2D(Z,M,ZE) evaluates the potential both at Z and ZE at
%   the same time.
%
%   P = FMM2D(Z,M,...) or [...] = FMM2D(Z,M,ZE,...) allows options to 
%   be passed in the form of property/value-pairs. See the table below 
%   for available properties.
%
%   Property     Value/{Default}     Description
%   -----------------------------------------------------------------------
%   pot          {0} | 1             0: Logarithmic potential,
%                                    P = -M*LOG(ABS(Z)).
%                                    1: Harmonic (doublet) potential,
%                                    P = -M/Z.
%
%   tol          Scalar {1e-6}       Tolerance. The choice tol = 0 means  
%                                    a direct evaluation.
%
%   ndirect      Scalar integer {5}  The number of particles in each
%                                    box at the finest level.
%
%   smooth       {'dirac'} |         Near-field models, see below.
%                'rankine' |
%                'scully' |
%                'oseen'
%
%   xopt         Scalar {1e-3}       Distance to the extreme point of the
%                                    smoother.
%
%   cutoff       Scalar {2e-3}       Cutoff-radius of the smoother.
%
%   cont         {true} | false      Enforce continuity to match the near- 
%                                    and far-fields.
%
%   out          {'sol'} | 'mesh'    Output solution/mesh. Type
%                                    'type fmm2d' for more information.
%
%   The precise definitions of the smoothers as defined via the properties
%   {'smooth','xopt','cutoff','cont'} are quite complicated. For this
%   purpose, reference implementations have been prepared. See the files
%   LOGPOT.M and INVPOT.M for details.
%
%   Note, however, that the cut-off radius of the smoother is only
%   respected in direct near-field computations. If this radius is
%   large compared to the boxes used by the algorithm in resolving the
%   computational domain, then the reference implementations in
%   LOGPOT/INVPOT are incorrect since they respect the cut-off radius
%   globally.
%
%   Example:
%     % N random masses
%     N = 20;
%     z = complex(rand(N,1),rand(N,1));
%     m = rand(N,1);
%
%     % mesh
%     [X,Y] = meshgrid(linspace(0,1,40));
%     e = complex(X,Y);
%
%     % direct evaluation of the Logarithmic potential with Oseen smoothing
%     p = fmm2d(z,m,e,'tol',0,'smooth','oseen');
%     contour(X,Y,p);
%     hold on, plot(real(z),imag(z),'ko');
%
%   See also LOGPOT, INVPOT.

% S. Engblom 2009-10-12 (Revision, performance)
% S. Engblom 2009-04-16 (Major revision, adaptivity)
% S. Engblom 2007-07-08 (Major revision)
% S. Engblom 2007-05-15 (Revision)
% S. Engblom 2006-07-10
% Based on the code 'fmm2dlp' by the same author.

% Notes:
%   -The call P = FMM2D(Z,M,Z,...) is interpreted as P = FMM2D(Z,M,...)
%   through comparision of pointers.
%
%   -The syntax Y = FMM2D(X) computes Y = EXPINT(X) very rapidly. This
%   syntax is intended for internal testing only.
%
%   -The syntax M = FMM2D(...,'out','mesh') returns the FMM mesh in a simple
%   format (see below and see also MESHPLOT for an example). This syntax is
%   mainly intended for debugging and pedagogical use.
%
%   [M,C] = FMM2D(...,'out','mesh') additionally returns cell vectors M and
%   C containing mesh- and sparse connection matrices.
%
%   M{l} for l = 1..nlevels contains 2-by-Nl complex matrices where Nl =
%   4^(l-1) is the number of boxes at level l. M{l}(1,:) are center points
%   for each rectangle and M{l}(2,:) contains the corresponding radii
%   (i.e. distances from centers to upper right corners).
%
%   C{l}(i,j) = 1 if box i and j are strongly connected (interact directly),
%   and C{l}(i,j) = -1 whenever box i and j interact through cluster to
%   cluster interactions at level l. If all(C{l}(i,j+1:4) = -1), then a
%   cluster to cluster interaction via the parent of j is used and
%   C{l}(i,j+1:4) = -2 in order to specify this. At the final level,
%   C{nlevel}(i,j) = 2 if box i and j interact through cluster to particle
%   (or particle to cluster) interactions. Only the strictly lower
%   triangular part of the matrix is stored. If required, C{i} =
%   speye(size(C{i}))+C{i}''+C{i}' yields the complete matrix.
%
%   See also MESHPLOT, MESHSTATS, MESHQUAL.

error('.MEX-file not found on path.');
