/* direct.c */

/* S. Engblom 2010-02-18 */

#include "expint.h"
#include "direct.h"

// code responds to #define FMM_INLINECOMPLEX_DIRECT (see fmm.c)

void rlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void rlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);

void rinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont);
void rinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);
void zinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont);

/*------------------------------------------------------------------------*/
void directInteract(int N,
                    const double *zr,const double *zi,
		    const double *mr,const double *mi,
		    int NE,
		    const double *er,const double *ei,
                    double *pr,double *pi,
		    double *qr,double *qi,int pot,
		    SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Direct evaluation of the potential from N pointmasses (mr,mi) (with
   mi possibly NULL) at the positions (zr,zi). The evaluation is done
   at the points (zr,zi) if (pr,pi) is not NULL and/or at the points
   (er,ei) if (qr,qi) is not NULL. The result is stored in the vectors
   (pr,pi) and (qr,qi), respectively, which must be allocated and
   cleared prior to call. Logarithmic potential for pot == 0, harmonic
   potential otherwise. The smoothing is done using smoother smooth
   with parameters (xopt,cutoff,cont). */
{
  // evaluation at (zr,zi)
  if (pr != NULL) {
    // mass is real
    if (mi == NULL)
      pot ? 
	rinv2(N,zr,zi,mr,pr,pi,smooth,xopt,cutoff,cont) : 
	rlog2(N,zr,zi,mr,pr,pi,smooth,xopt,cutoff,cont);
    // mass is complex
    else
      pot ? 
	zinv2(N,zr,zi,mr,mi,pr,pi,smooth,xopt,cutoff,cont) : 
	zlog2(N,zr,zi,mr,mi,pr,pi,smooth,xopt,cutoff,cont);
  }

  // evaluation at (er,ei)
  if (qr != NULL) {
    // mass is real
    if (mi == NULL)
      pot ? 
	rinv(N,zr,zi,mr,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont) : 
	rlog(N,zr,zi,mr,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont);
    // mass is complex
    else
      pot ? 
	zinv(N,zr,zi,mr,mi,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont) :
	zlog(N,zr,zi,mr,mi,NE,er,ei,qr,qi,smooth,xopt,cutoff,cont);
  }
}
/*------------------------------------------------------------------------*/
/* Four routines for the logarithmic potential. */
void rlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Logarithmic potential for N potentials located at (zr,zi) with real
   mass mr. The contribution is added to (pr,pi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 0.5*log(re*re+im*im);
#else
	const double P = log(hypot(zr[i]-zr[j],zi[i]-zi[j]));
#endif
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
#else
    cutoff = xopt;
    shape = log(cutoff)-0.5;
#endif
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*shape;
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
#else
	const double ad = hypot(zr[i]-zr[j],zi[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
#else
    shape = xopt*xopt;
    if (cont)
      scale = 2.0*log(cutoff)/log(shape+cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*0.5*log(shape)*scale;
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
#else
	const double ad = hypot(zr[i]-zr[j],zi[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*log(shape+ad*ad)*scale;
#endif
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+0.5*expint(shape*cutoff*cutoff));
#endif
    for (int i = 0; i < N; i++) {
      pr[i] += mr[i]*0.5*(log(shape)+gamma_)*scale;
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#else
	const double ad = hypot(zr[i]-zr[j],zi[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else if (ad != 0.0)
	  P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#endif
	pr[i] -= mr[j]*P;
	pr[j] -= mr[i]*P;
      }
    }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zlog2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Same as above but with complex mass (mr,mi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 0.5*log(re*re+im*im);
#else
	const double P = log(hypot(zr[i]-zr[j],zi[i]-zi[j]));
#endif
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
#else
    cutoff = xopt;
    shape = log(cutoff)-0.5;
#endif
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*shape;
      pi[i] -= mi[i]*shape;
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
#else
	const double ad = hypot(zr[i]-zr[j],zi[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
#else
    shape = xopt*xopt;
    if (cont)
      scale = 2.0*log(cutoff)/log(shape+cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++) {
      pr[i] -= mr[i]*0.5*log(shape)*scale;
      pi[i] -= mi[i]*0.5*log(shape)*scale;
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
#else
	const double ad = hypot(zr[i]-zr[j],zi[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*log(shape+ad*ad)*scale;
#endif
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+0.5*expint(shape*cutoff*cutoff));
#endif
    for (int i = 0; i < N; i++) {
      pr[i] += mr[i]*0.5*(log(shape)+gamma_)*scale;
      pi[i] += mi[i]*0.5*(log(shape)+gamma_)*scale;
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#else
	const double ad = hypot(zr[i]-zr[j],zi[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else if (ad != 0.0)
	  P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#endif
	pr[i] -= mr[j]*P;
	pi[i] -= mi[j]*P;
	pr[j] -= mr[i]*P;
	pi[j] -= mi[i]*P;
      }
    }
    break;
  }
}
/*------------------------------------------------------------------------*/
void rlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Logarithmic potential evaluated at NE points (er,ei) with
   potentials at the N points (zr,zi) with real mass mr. The
   contribution is added to (qr,qi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 0.5*log(re*re+im*im);
#else
	  const double P = log(hypot(er[i]-zr[j],ei[i]-zi[j]));
#endif
	  qr[i] -= mr[j]*P;
	}
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
#else
    cutoff = xopt;
    shape = log(cutoff)-0.5;
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
#else
	const double ad = hypot(er[i]-zr[j],ei[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	qr[i] -= mr[j]*P;
      }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
#else
    shape = xopt*xopt;
    if (cont)
      scale = 2.0*log(cutoff)/log(shape+cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
#else
	const double ad = hypot(er[i]-zr[j],ei[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*log(shape+ad*ad)*scale;
#endif
	qr[i] -= mr[j]*P;
      }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+0.5*expint(shape*cutoff*cutoff));
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#else
	const double ad = hypot(er[i]-zr[j],ei[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else if (ad != 0.0)
	  P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#endif
	qr[i] -= mr[j]*P;
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zlog(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Same as above but with complex mass (mr,mi). */
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 0.5*log(re*re+im*im);
#else
	  const double P = log(hypot(er[i]-zr[j],ei[i]-zi[j]));
#endif
	  qr[i] -= mr[j]*P;
	  qi[i] -= mi[j]*P;
	}
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 0.5*log(cutoff)-0.5;
#else
    cutoff = xopt;
    shape = log(cutoff)-0.5;
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*(ad/cutoff)+shape;
#else
	const double ad = hypot(er[i]-zr[j],ei[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	qr[i] -= mr[j]*P;
	qi[i] -= mi[j]*P;
      }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = log(cutoff)/log(shape+cutoff);
#else
    shape = xopt*xopt;
    if (cont)
      scale = 2.0*log(cutoff)/log(shape+cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else
	  P = 0.5*log(shape+ad)*scale;
#else
	const double ad = hypot(er[i]-zr[j],ei[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else
	  P = 0.5*log(shape+ad*ad)*scale;
#endif
	qr[i] -= mr[j]*P;
	qi[i] -= mi[j]*P;
      }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+expint(shape*cutoff));
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = log(cutoff)/(log(cutoff)+0.5*expint(shape*cutoff*cutoff));
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 0.5*log(ad);
	else if (ad != 0.0)
	  P = 0.5*(log(ad)+expint(shape*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#else
	const double ad = hypot(er[i]-zr[j],ei[i]-zi[j]);
	double P;
	if (ad >= cutoff)
	  P = log(ad);
	else if (ad != 0.0)
	  P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	else
	  P = -0.5*(log(shape)+gamma_)*scale;
#endif
	qr[i] -= mr[j]*P;
	qi[i] -= mi[j]*P;
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
/* Now follows the 4 corresponding functions for the Harmonic
   potential. */
void rinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 1.0/(re*re+im*im);
	pr[i] -= mr[j]*P*re;
	pi[i] += mr[j]*P*im;
	pr[j] += mr[i]*P*re;
	pi[j] -= mr[i]*P*im;
#else
	const dcmplx P = 1.0/(zr[i]-zr[j]+I*(zi[i]-zi[j]));
	pr[i] -= mr[j]*creal(P);
	pi[i] -= mr[j]*cimag(P);
	pr[j] += mr[i]*creal(P);
	pi[j] += mr[i]*cimag(P);
#endif
      }
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
#else
    cutoff = xopt;
    shape = 1.0/(cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	pr[i] -= mr[j]*P*re;
	pi[i] += mr[j]*P*im;
	pr[j] += mr[i]*P*re;
	pi[j] -= mr[i]*P*im;
#else
	const dcmplx d = zr[i]-zr[j]+I*(zi[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)*shape;
	pr[i] -= mr[j]*creal(P);
	pi[i] -= mr[j]*cimag(P);
	pr[j] += mr[i]*creal(P);
	pi[j] += mr[i]*cimag(P);
#endif
      }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
#else
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/(cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	pr[i] -= mr[j]*P*re;
	pi[i] += mr[j]*P*im;
	pr[j] += mr[i]*P*re;
	pi[j] -= mr[i]*P*im;
#else
	const dcmplx d = zr[i]-zr[j]+I*(zi[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)/(shape+ad*ad)*scale;
	pr[i] -= mr[j]*creal(P);
	pi[i] -= mr[j]*cimag(P);
	pr[j] += mr[i]*creal(P);
	pi[j] += mr[i]*cimag(P);
#endif
      }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  pr[i] -= mr[j]*P*re;
	  pi[i] += mr[j]*P*im;
	  pr[j] += mr[i]*P*re;
	  pi[j] -= mr[i]*P*im;
	}
#else
	const dcmplx d = zr[i]-zr[j]+I*(zi[i]-zi[j]);
	const double ad = cabs(d);
	if (ad != 0.0) {
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = -expm1(-shape*ad*ad)/d*scale;
	  pr[i] -= mr[j]*creal(P);
	  pi[i] -= mr[j]*cimag(P);
	  pr[j] += mr[i]*creal(P);
	  pi[j] += mr[i]*cimag(P);
	}
#endif
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zinv2(int N,
	   const double *zr,const double *zi,
	   const double *mr,const double *mi,
	   double *pr,double *pi,
	   SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double P = 1.0/(re*re+im*im);
	pr[i] -= (mr[j]*re+mi[j]*im)*P;
	pi[i] += (mr[j]*im-mi[j]*re)*P;
	pr[j] += (mr[i]*re+mi[i]*im)*P;
	pi[j] -= (mr[i]*im-mi[i]*re)*P;
#else
	const dcmplx P = 1.0/(zr[i]-zr[j]+I*(zi[i]-zi[j]));
	pr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	pi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
	pr[j] += mr[i]*creal(P)-mi[i]*cimag(P);
	pi[j] += mr[i]*cimag(P)+mi[i]*creal(P);
#endif
      }
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
#else
    cutoff = xopt;
    shape = 1.0/(cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	pr[i] -= (mr[j]*re+mi[j]*im)*P;
	pi[i] += (mr[j]*im-mi[j]*re)*P;
	pr[j] += (mr[i]*re+mi[i]*im)*P;
	pi[j] -= (mr[i]*im-mi[i]*re)*P;
#else
	const dcmplx d = zr[i]-zr[j]+I*(zi[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)*shape;
	pr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	pi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
	pr[j] += mr[i]*creal(P)-mi[i]*cimag(P);
	pi[j] += mr[i]*cimag(P)+mi[i]*creal(P);
#endif
      }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
#else
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/(cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	pr[i] -= (mr[j]*re+mi[j]*im)*P;
	pi[i] += (mr[j]*im-mi[j]*re)*P;
	pr[j] += (mr[i]*re+mi[i]*im)*P;
	pi[j] -= (mr[i]*im-mi[i]*re)*P;
#else
	const dcmplx d = zr[i]-zr[j]+I*(zi[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)/(shape+ad*ad)*scale;
	pr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	pi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
	pr[j] += mr[i]*creal(P)-mi[i]*cimag(P);
	pi[j] += mr[i]*cimag(P)+mi[i]*creal(P);
#endif
      }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff*cutoff);
#endif
    for (int i = 0; i < N; i++)
      for (int j = i+1; j < N; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = zr[i]-zr[j],im = zi[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  pr[i] -= (mr[j]*re+mi[j]*im)*P;
	  pi[i] += (mr[j]*im-mi[j]*re)*P;
	  pr[j] += (mr[i]*re+mi[i]*im)*P;
	  pi[j] -= (mr[i]*im-mi[i]*re)*P;
	}
#else
	const dcmplx d = zr[i]-zr[j]+I*(zi[i]-zi[j]);
	const double ad = cabs(d);
	if (ad != 0.0) {
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = -expm1(-shape*ad*ad)/d*scale;
	  pr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	  pi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
	  pr[j] += mr[i]*creal(P)-mi[i]*cimag(P);
	  pi[j] += mr[i]*cimag(P)+mi[i]*creal(P);
	}
#endif
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void rinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 1.0/(re*re+im*im);
	  qr[i] -= mr[j]*P*re;
	  qi[i] += mr[j]*P*im;
#else
	  const dcmplx P = 1.0/(er[i]-zr[j]+I*(ei[i]-zi[j]));
	  qr[i] -= mr[j]*creal(P);
	  qi[i] -= mr[j]*cimag(P);
#endif
	}
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
#else
    cutoff = xopt;
    shape = 1.0/(cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	qr[i] -= mr[j]*P*re;
	qi[i] += mr[j]*P*im;
#else
	const dcmplx d = er[i]-zr[j]+I*(ei[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)*shape;
	qr[i] -= mr[j]*creal(P);
	qi[i] -= mr[j]*cimag(P);
#endif
      }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
#else
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/(cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	qr[i] -= mr[j]*P*re;
	qi[i] += mr[j]*P*im;
#else
	const dcmplx d = er[i]-zr[j]+I*(ei[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)/(shape+ad*ad)*scale;
	qr[i] -= mr[j]*creal(P);
	qi[i] -= mr[j]*cimag(P);
#endif
      }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  qr[i] -= mr[j]*P*re;
	  qi[i] += mr[j]*P*im;
	}
#else
	const dcmplx d = er[i]-zr[j]+I*(ei[i]-zi[j]);
	const double ad = cabs(d);
	if (ad != 0.0) {
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = -expm1(-shape*ad*ad)/d*scale;
	  qr[i] -= mr[j]*creal(P);
	  qi[i] -= mr[j]*cimag(P);
	}
#endif
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
void zinv(int N,
	  const double *zr,const double *zi,
	  const double *mr,const double *mi,
	  int NE,
	  const double *er,const double *ei,
	  double *qr,double *qi,
	  SMOOTHER smooth,double xopt,double cutoff,bool cont)
{
  double shape,scale = 1.0;

  switch (smooth) {
  case DIRAC:
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++)
	if (er[i] != zr[j] || ei[i] != zi[j]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	  const double re = er[i]-zr[j],im = ei[i]-zi[j];
	  const double P = 1.0/(re*re+im*im);
	  qr[i] -= (mr[j]*re+mi[j]*im)*P;
	  qi[i] += (mr[j]*im-mi[j]*re)*P;
#else
	  const dcmplx P = 1.0/(er[i]-zr[j]+I*(ei[i]-zi[j]));
	  qr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	  qi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
#endif
	}
    break;

  case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = xopt*xopt;
    shape = 1.0/cutoff;
#else
    cutoff = xopt;
    shape = 1.0/(cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = shape;
	qr[i] -= (mr[j]*re+mi[j]*im)*P;
	qi[i] += (mr[j]*im-mi[j]*re)*P;
#else
	const dcmplx d = er[i]-zr[j]+I*(ei[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)*shape;
	qr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	qi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
#endif
      }
    break;

  case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/cutoff;
#else
    shape = xopt*xopt;
    if (cont)
      scale = 1.0+shape/(cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad >= cutoff)
	  P = 1.0/ad;
	else
	  P = 1.0/(shape+ad)*scale;
	qr[i] -= (mr[j]*re+mi[j]*im)*P;
	qi[i] += (mr[j]*im-mi[j]*re)*P;
#else
	const dcmplx d = er[i]-zr[j]+I*(ei[i]-zi[j]);
	const double ad = cabs(d);
	dcmplx P;
	if (ad >= cutoff)
	  P = 1.0/d;
	else
	  P = conj(d)/(shape+ad*ad)*scale;
	qr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	qi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
#endif
      }
    break;

  case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    cutoff = cutoff*cutoff;
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff);
#else
    shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      scale = -1.0/expm1(-shape*cutoff*cutoff);
#endif
    for (int j = 0; j < N; j++)
      for (int i = 0; i < NE; i++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	const double re = er[i]-zr[j],im = ei[i]-zi[j];
	const double ad = re*re+im*im;
	double P;
	if (ad != 0.0) {
	  if (ad >= cutoff)
	    P = 1.0/ad;
	  else
	    P = -expm1(-shape*ad)/ad*scale;
	  qr[i] -= (mr[j]*re+mi[j]*im)*P;
	  qi[i] += (mr[j]*im-mi[j]*re)*P;
	}
#else
	const dcmplx d = er[i]-zr[j]+I*(ei[i]-zi[j]);
	const double ad = cabs(d);
	if (ad != 0.0) {
	  dcmplx P;
	  if (ad >= cutoff)
	    P = 1.0/d;
	  else
	    P = -expm1(-shape*ad*ad)/d*scale;
	  qr[i] -= mr[j]*creal(P)-mi[j]*cimag(P);
	  qi[i] -= mr[j]*cimag(P)+mi[j]*creal(P);
	}
#endif
      }
    break;
  }
}
/*------------------------------------------------------------------------*/
