/* fmm.c */

/* S. Engblom 2010-02-18 */

#include "fmm.h"
#include "fmmshift.h"
#include "expint.h"

/* #defines: (preferably handled via make)
   ---------------------------------------
   FMM_INLINECOMPLEX
                 Avoid complex arithmetic in various places in the
		 FMM-algorithm and use inline versions instead. Care
		 has been taken to do this in such a way that the
		 result does not differ from C99's native complex
		 aritmetic. When in doubt, try #undefining
		 FMM_INLINECOMPLEX.

   FMM_INLINECOMPLEX_DIRECT
                 Inline complex aritmetic in direct evaluations. On
		 some platforms, it is faster by a factor of up to 8
		 not to rely on certain math-functions and/or complex
		 arithmetics, but to write the complex arithmetic
		 inline. On the one hand this implies possible
		 overflow as the evaluations of complex division and
		 absolute values demands properly scaling; on the
		 other hand this will only be observed with very
		 extreme scaling of the input since the operations are
		 only used in near-field computations. Again, when in
		 doubt, try #undefining FMM_INLINECOMPLEX_DIRECT.

   FMM_SHRINK    Postsplit shrinking of boxes on/off: this option is
		 kept here for the purpose of visual
		 demonstrations. Also good to have this option if
		 time-dependent FMM is implemented.

   It is recommended to #define all these.   
*/

// high-level routines
void MPexp_setup(MPexp *this,int N,int NE,
		 const double *zr,const double *zi,
		 const double *mr,const double *mi,
		 const double *er,const double *ei,
		 int pot,double tol,int Ndirect,double eta,bool meshonly);
void MPexp_free(MPexp *this);
void MPexp_eval2(const MPexp *this,
		 double *pr,double *pi,
		 const double *zr,const double *zi,
		 const double *mr,const double *mi,
		 int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont);
void MPexp_eval(const MPexp *this,
		double *qr,double *qi,
		const double *er,const double *ei,
		const double *zr,const double *zi,
		const double *mr,const double *mi,
		int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont);

// helper functions
void MPexp_minmax_(double *xmin,double *xmax,double *ymin,double *ymax,
		   int N,const double *x,const double *y);
void MPexp_box_(MPexp *this,int N,int NE,
		const double *zr,const double *zi,
		const double *er,const double *ei);
void MPexp_partition_(int begin,int *im,int end,double *z0,
		      double *min,double *max,
		      int *ix,const double *z,int Nmax,double eta);
void MPexp_split_(int begin,int *im,int end,double z0,
		  double *min,double *max,
		  int *ix,const double *z);
void MPexp_sort_(MPexp *this,
		 int N,const double *zr,const double *zi,
		 int NE,const double *er,const double *ei,
		 int *maxm2p,int *levm2p,int Ndirect,double eta);
void MPexp_smooth_(int pot,SMOOTHER smooth,double xopt,bool cont,
		   double *cutoff,double *shape,double *scale);

// low-level routines
void mpexp_init(mpexp *this,int p,
		const double zr,const double zi,
		const double mr,const double mi,
		int pot);
void mpexp_initp(mpexp *this,int p,
		 const double zr,const double zi,
		 const double mr,const double mi,
		 int pot);
bool mpexp_theta(const mpexp *this,const mpexp *that);
int mpexp_theta2(const mpexp *this,const mpexp *that);
void mpexp_eval(const mpexp *this,int p,
		double *vr,double *vi,
		const double *zr,const double *zi,
		const int *ix,int begin,int end);
void mpexp_evalmp(const mpexp *this,int p,int pot,
		  double *vr,double *vi,
		  const double *zr,const double *zi,
		  const int *ix,int begin,int end);
void mpexp_directInteract2(double *pr,double *pi,
			   const double *zr,const double *zi,
			   const double *mr,const double *mi,
			   const int *ix,
			   int begin1,int end1,int begin2,int end2,
			   int pot,SMOOTHER smooth,
			   double cutoff,double shape,double scale);
void mpexp_directInteract(double *qr,double *qi,
			  const double *er,const double *ei,
			  const int *jx,int begin1,int end1,
			  const double *zr,const double *zi,
			  const double *mr,const double *mi,
			  const int *ix,int begin2,int end2,
			  int pot,SMOOTHER smooth,
			  double cutoff,double shape,double scale);

/*------------------------------------------------------------------------*/
void fmm2dInteract(int N,
		   const double *zr,const double *zi,
		   const double *mr,const double *mi,
		   int NE,
		   const double *er,const double *ei,
		   double *pr,double *pi,
		   double *qr,double *qi,
		   int pot,double tol,int Ndirect,double eta,
		   SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation by the fast multipole method to accuracy TOL of the
   potential from N pointmasses (mr,mi) (with mi possibly NULL) at the
   positions (zr,zi). The evaluation is done at the points (zr,zi) if
   (pr,pi) is not NULL and/or at the points (er,ei) if (qr,qi) is not
   NULL (it is required that NE is zeroe if (qr,qi) are NULL). The
   result is added to the vectors (pr,pi) and (qr,qi), respectively,
   which must be allocated prior to call. Logarithmic potential for
   pot == 0, harmonic potential otherwise. The smoothing is done using
   smoother smooth with parameters (xopt,cutoff,cont). */
{
  // early return: cannot allocate anything
  if (N <= 0) return;

  // multipole structure
  MPexp MP;

  // allocate and setup representation
  MPexp_setup(&MP,N,NE,zr,zi,mr,mi,er,ei,pot,tol,Ndirect,eta,false);

  StartTime;
  // evaluation at potentials (zr,zi)
  if (pr != NULL) MPexp_eval2(&MP,pr,pi,zr,zi,mr,mi,
			      pot,smooth,xopt,cutoff,cont);

  // evaluation at coordinates (er,ei)
  if (qr != NULL) MPexp_eval(&MP,qr,qi,er,ei,zr,zi,mr,mi,
			     pot,smooth,xopt,cutoff,cont);
  StopTime;
  PrintTime("Evaluation: %f\n");

  // dellocate
  MPexp_free(&MP);
}
/*------------------------------------------------------------------------*/
void fmm2dGetMesh(int nlhs,mxArray *plhs[],
		  int N,
		  const double *zr,const double *zi,
		  const double *mr,const double *mi,
		  int NE,
		  const double *er,const double *ei,
		  int pot,double tol,int Ndirect,double eta)
/* Returns the multipole mesh in a cell vector of complex
   mxArrays. Additionally, all sparse connection matrices are
   optionally returned. For debugging and pedagogical purposes. */
{
  if (N <= 0) return;

  // setup multipole tree
  MPexp MP;
  MPexp_setup(&MP,N,NE,zr,zi,mr,mi,er,ei,pot,tol,Ndirect,eta,true);

  // create cell vector and copy the FMM mesh
  plhs[0] = mxCreateCellMatrix(1,MP.nlevel+1);
  for (int i = 0; i <= MP.nlevel; i++) {
    mxArray *cell = mxCreateDoubleMatrix(2,MP.lptr[i+1]-MP.lptr[i],
					 mxCOMPLEX);
    double *mshr = mxGetPr(cell), *mshi = mxGetPi(cell);
    mpexp *offset = &MP.root[MP.lptr[i]];

    for (int j = 0; j < MP.lptr[i+1]-MP.lptr[i]; j++) {
      mshr[2*j] = creal(offset[j].z0);
      mshi[2*j] = cimag(offset[j].z0);
      mshr[2*j+1] = creal(offset[j].d0);
      mshi[2*j+1] = cimag(offset[j].d0);
    }
    mxSetCell(plhs[0],i,cell);
  }

  // additionally, the connection matrices
  if (nlhs > 1) {
    plhs[1] = mxCreateCellMatrix(1,MP.nlevel+1);
    for (int i = 0; i <= MP.nlevel; i++) {
      const mpSparse *S = &MP.connect[i];
      const int N = MP.lptr[i+1]-MP.lptr[i];
      mxArray *cell = mxCreateSparse(N,N,S->jcptr[N],mxREAL);
      mwIndex *jc = mxGetJc(cell),*ir = mxGetIr(cell);
      double *pr = mxGetPr(cell);
      // mwIndex not defined? use '-DmwIndex=int' in the mex-command!

      jc[0] = 0;
      for (int j = 0,k; j < N; j++) {
	jc[j+1] = S->jcptr[j+1];

	// near connections
	if (i < MP.nlevel)
	  for (k = jc[j]; k < S->kcptr[j]; k++) {
	    ir[k] = S->ir[k];
	    pr[k] = 1.0;
	  }
	else {
	  // final level contains more information:
	  for (k = jc[j]; k < S->kcptr[j+N]; k++) {
	     ir[k] = S->ir[k];
	     pr[k] = 1.0;
	  }
	  for ( ; k < S->kcptr[j]; k++) {
	    ir[k] = abs(S->ir[k]); // clear sign
	    pr[k] = 2.0; // near-far connection
	  }
	}
	// far connections
	for ( ; k < S->jcptr[j+1]; k++)
	  if (S->ir[k] < 0) { // far connection between levels
	    ir[k] = S->ir[k+1]+1;
	    pr[k] = -2.0;
	    for (int d = 1; d < 4; d++,k++) {
	      ir[k+1] = S->ir[k+1];
	      pr[k+1] = pr[k];
	    }
	  }
	  else {
	    ir[k] = S->ir[k];
	    pr[k] = -1.0;
	  }
      }
      mxSetCell(plhs[1],i,cell);
    }
  }

  // deallocate and return
  MPexp_free(&MP);
}
/*------------------------------------------------------------------------*/
void MPexp_setup(MPexp *this,int N,int NE,
                 const double *zr,const double *zi,
                 const double *mr,const double *mi,
                 const double *er,const double *ei,
		 int pot,double tol,int Ndirect,double eta,bool meshonly)
/* Setups and allocates a multipole tree for N potentials at (zr,zi)
   to be evaluated at the NE points (er,ei) to tolerance tol. The
   discretization is determined so that the number of potentials in
   each box at the finest level is <= Ndirect. If the boolean variable
   meshonly is true, then only the multipole tree and the connectivity
   information is determined. */
{
  StartTime;

  // size of the domain
  MPexp_box_(this,N,NE,zr,zi,er,ei);

  // number of multipole terms
  this->pcoeff = (int)floor(log2(4.0*tol)/log2(theta)-1.0)+(pot != 0);
  /* the magical constants 4 is a (not overly optimistic) estimate
     obtained from numerous experiments with the model error ~
     constant * theta^(p+1) and theta = 0.5 */
  if (this->pcoeff < 1) this->pcoeff = 1;

  // number of levels
  this->nlevel = ceil(0.5*log2(5.0/8.0*(double)N/Ndirect));
  if (this->nlevel < 0) this->nlevel = 0;
  /* the magical constant 5/8 comes from solving for the number of
     levels the equation E(N/Nf) = Ndirect, where E is expectation, N
     is considered a uniform random number on [1,M] (with M tending to
     infinity) and Nf is defined below */

  /* used frequently below: total number of coefficients, number of
     boxes at the finest level and total number of boxes */
  const int P = this->pcoeff+1;
  const int Nf = 1 << (this->nlevel << 1);
  const int Nt = (Nf-1)/3+Nf;

  // pointer to levels
  this->lptr = mxMalloc((this->nlevel+2)*sizeof(int));
  this->lptr[0] = 0;
  for (int level = 0,nb = 1; level <= this->nlevel; level++, nb <<= 2)
    this->lptr[level+1] = this->lptr[level]+nb;

  // multipole tree
  this->root = mxMalloc(Nt*sizeof(mpexp));

  // multipole coefficients
  dcmplx *coeffs1 = mxCalloc(P*Nt,sizeof(dcmplx));
  dcmplx *coeffs2 = mxCalloc(P*Nt,sizeof(dcmplx));
  for (int j = 0; j < Nt; j++) {
    this->root[j].coeff1 = coeffs1; coeffs1 += P;
    this->root[j].coeff2 = coeffs2; coeffs2 += P;
  }

  // permutation of input
  this->ixptr = mxMalloc((Nf+1)*sizeof(int));
  this->ix = mxMalloc(N*sizeof(int));
  if (NE) {
    this->jxptr = mxMalloc((Nf+1)*sizeof(int));
    this->jx = mxMalloc(NE*sizeof(int));
  }
  else
    this->jxptr = this->jx = NULL;

  // connections between boxes at each level
  this->connect = mxMalloc((this->nlevel+1)*sizeof(mpSparse));
  StopTime;
  PrintTime("Alloc: %f\n");

  StartTime;
  /* setup the multipole mesh, sort all potentials and evaluation
     points and determine all connections */
  int maxm2p,levm2p;
  MPexp_sort_(this,N,zr,zi,NE,er,ei,&maxm2p,&levm2p,Ndirect,eta);
  StopTime;
  PrintTime("Sort: %f\n");
  if (meshonly) return;

  /* Start of the actual multipole algorithm */

  /* Step #1: Initialization at the finest level. */
  StartTime;
  {
    mpexp *offset = &this->root[this->lptr[this->nlevel]];
    mpSparse *C = &this->connect[this->nlevel];

    if (mi == NULL)
      for (int j = 0; j < Nf; j++) {
	// initialize multipole expansion in box j
	for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
	  mpexp_init(&offset[j],this->pcoeff,
		     zr[this->ix[i]],zi[this->ix[i]],
		     mr[this->ix[i]],0.0,pot);

	// 'less near' connections: shift into polynomial directly
	for (int jj = C->kcptr[j+Nf]; jj < C->kcptr[j]; jj++) {
	  const int k = C->ir[jj];
	  if (k < 0)
	    for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
	      mpexp_initp(&offset[-k],this->pcoeff,
			  zr[this->ix[i]],zi[this->ix[i]],
			  mr[this->ix[i]],0.0,pot);
	  else
	    for (int i = this->ixptr[k]; i < this->ixptr[k+1]; i++)
	      mpexp_initp(&offset[j],this->pcoeff,
			  zr[this->ix[i]],zi[this->ix[i]],
			  mr[this->ix[i]],0.0,pot);
	}
      }
    else
      for (int j = 0; j < Nf; j++) {
	for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
	  mpexp_init(&offset[j],this->pcoeff,
		     zr[this->ix[i]],zi[this->ix[i]],
		     mr[this->ix[i]],mi[this->ix[i]],pot);

	for (int jj = C->kcptr[j+Nf]; jj < C->kcptr[j]; jj++) {
	  const int k = C->ir[jj];
	  if (k < 0)
	    for (int i = this->ixptr[j]; i < this->ixptr[j+1]; i++)
	      mpexp_initp(&offset[-k],this->pcoeff,
			  zr[this->ix[i]],zi[this->ix[i]],
			  mr[this->ix[i]],mi[this->ix[i]],pot);
	  else
	    for (int i = this->ixptr[k]; i < this->ixptr[k+1]; i++)
	      mpexp_initp(&offset[j],this->pcoeff,
			  zr[this->ix[i]],zi[this->ix[i]],
			  mr[this->ix[i]],mi[this->ix[i]],pot);
	}
      }
  }
  StopTime;
  PrintTime("Step #1: %f\n");

  /* Step #2: Upward pass of the fast multipole algorithm. After the
     finest level has been initialized, this loop performs the
     'upward' pass where all multipole contributions are shifted and
     added onto all coarser levels. */
  StartTime;
  shift_alloc(this->pcoeff,pot,maxm2p); // prepare to shift
  for (int l = this->nlevel,nb = Nf; l > 0; l--, nb >>= 2) {
    mpexp *parent = &this->root[this->lptr[l-1]];
    mpexp *child = &this->root[this->lptr[l]];

    // loop over all children at this level
    for (int j = 0; j < nb; j += 4) {
      // shift and add result from 4 children at once
      shift_m2m(child[j].z0-parent->z0,child[j+1].z0-parent->z0,
		child[j+2].z0-parent->z0,child[j+3].z0-parent->z0,
		parent->coeff1,child[j].coeff1);
      parent++;
    }
  }
  StopTime;
  PrintTime("Step #2: %f\n");

  /* Step #3: Downward pass of the fast multipole method. This shifts
     expansions from the interaction list of each box to the center of
     the box and adds up the result. Before continuing to finer
     levels, each parent shifts and adds its expansion to each of its
     children. */
  StartTime;
  for (int l = levm2p-1,nb = 1 << 2*levm2p; l < this->nlevel; l++, nb <<= 2) {
    mpexp *parent = &this->root[this->lptr[l]];
    mpexp *child = &this->root[this->lptr[l+1]];
    mpSparse *C = &this->connect[l+1];

    /* tricky part: loop over all boxes (parents) at this level and
       symmetrically shift results to and from the interaction list of
       each child */
    for (int j = 0; j < nb; ) {
      // contribution from parent to 4 children at once
      shift_p2p(parent->z0-child[j].z0,parent->z0-child[j+1].z0,
		parent->z0-child[j+2].z0,parent->z0-child[j+3].z0,
		child[j].coeff2,parent->coeff2);
      parent++;

      // loop through the interaction list of each child
      for (int c = 0; c < 4; c++,j++)
	shift_m2ps(child[j].z0,&child[0].z0,sizeof(mpexp),
		   C->ir,C->kcptr[j],C->jcptr[j+1],
		   child[j].coeff1,child[j].coeff2,
		   child[0].coeff1,child[0].coeff2);
    }
  }
  StopTime;
  PrintTime("Step #3: %f\n");
}
/*------------------------------------------------------------------------*/
void MPexp_free(MPexp *this)
/* Deallocates the whole tree. */
{
  for (int l = this->nlevel; l >= 0; l--)
    mxFree(this->connect[l].jcptr);
  mxFree(this->connect);
  mxFree(this->jx);
  mxFree(this->jxptr);
  mxFree(this->ix);
  mxFree(this->ixptr);
  mxFree(this->root->coeff2);
  mxFree(this->root->coeff1);
  mxFree(this->root);
  mxFree(this->lptr);
}
/*------------------------------------------------------------------------*/
void MPexp_minmax_(double *xmin,double *xmax,double *ymin,double *ymax,
		   int N,const double *x,const double *y)
/* Finds min/max of two vectors. Note: it is assumed that suitable
   initial values in xmin/xmax are already set. */
{
  // loop unrolled to get rid of a few N/2 comparisions
  for (int i = 0; i < N-1; i += 2) {
    double min,max;

    if (x[i] < x[i+1]) {
      min = x[i];
      max = x[i+1];
    }
    else {
      min = x[i+1];
      max = x[i];
    }
    if (min < *xmin) *xmin = min;
    if (max > *xmax) *xmax = max;

    if (y[i] < y[i+1]) {
      min = y[i];
      max = y[i+1];
    }
    else {
      min = y[i+1];
      max = y[i];
    }
    if (min < *ymin) *ymin = min;
    if (max > *ymax) *ymax = max;
  }
  if (N&1) {
    if (x[N-1] < *xmin)
      *xmin = x[N-1];
    else if (x[N-1] > *xmax)
      *xmax = x[N-1];

    if (y[N-1] < *ymin)
      *ymin = y[N-1];
    else if (y[N-1] > *ymax)
      *ymax = y[N-1];
  }
}
/*------------------------------------------------------------------------*/
void MPexp_box_(MPexp *this,int N,int NE,
		const double *zr,const double *zi,
		const double *er,const double *ei)
/* Determines and sets the dimensions of the computational domain. */
{
  this->xmin = zr[0];
  this->xmax = zr[0];
  this->ymin = zi[0];
  this->ymax = zi[0];
  MPexp_minmax_(&this->xmin,&this->xmax,
		&this->ymin,&this->ymax,
		N-1,zr+1,zi+1);

  // bounding box might become larger:
  if (NE) 
    MPexp_minmax_(&this->xmin,&this->xmax,
		  &this->ymin,&this->ymax,
		  NE,er,ei);
}
/*------------------------------------------------------------------------*/
#define SWAPI(X,Y) { int tmpi = X; X = Y; Y = tmpi; }

void MPexp_partition_(int begin,int *im,int end,double *z0,
		      double *min,double *max,
		      int *ix,const double *z,int Nmax,double eta)
/* Partitions the set of points z[ix[begin]] through z[ix[end]]
   (exclusive) in two sets indicated by [begin,im) and [im,end) such
   that all points are <= or >= than z0 (which is both input and
   output). Additionally, whenever min and max are nonzero, the
   minimum and maximum number are also determined. The division is
   done such that Nlarge <= Nmax, where Nlarge is the larger number of
   particles in the two lists. It is assumed that this split is
   possible and hence that Nmax is not too small.

   The routine is a workhorse and performance is important; for each
   box which is not a leaf it will be called three times. */
{
  int il,ir,left = begin,right = end;
  int Nsmall,Nlarge;

  // first split according to input z0 (geometric midpoint)
  const double z0_ = *z0; // compiler: this is really a constant!
  il = left-1; ir = right;
  do il++; while (il < right && z[ix[il]] < z0_);
  do ir--; while (left <= ir && z0_ < z[ix[ir]]);
  if (il < ir) {
    SWAPI(ix[il],ix[ir])
    /* after the first swap, there is no need to check for index out
       of bounds */
    for ( ; ; ) {
      do il++; while (z[ix[il]] < z0_);
      do ir--; while (z0_ < z[ix[ir]]);
      if (il >= ir) break;
      SWAPI(ix[il],ix[ir])
    }
  }

  // proceed with largest interval [left,il) or [il,right)
  Nsmall = il-begin;
  Nlarge = end-il;
  if (Nsmall > Nlarge) {
    Nlarge = Nsmall;
    *im = right = il;
  }
  else
    *im = left = il;
  const int im0 = *im; // saved for later

  while (Nlarge > Nmax) {
    if (right-left >= 3) {
      /* "median of 3": find pivot and create boundaries such that
	 left <= left+1 <= right-1 */
      const int mid = (left+right)>>1;
      SWAPI(ix[left+1],ix[mid])
      if (z[ix[left]] > z[ix[right-1]])
	SWAPI(ix[left],ix[right-1])
      if (z[ix[left+1]] > z[ix[right-1]])
        SWAPI(ix[left+1],ix[right-1])
      else if (z[ix[left]] > z[ix[left+1]])
	SWAPI(ix[left],ix[left+1])
      const double z1_ = *z0 = z[ix[left+1]];

      for (il = left+1,ir = right-1; ; ) {
	do il++; while (z[ix[il]] < z1_);
	do ir--; while (z1_ < z[ix[ir]]);
	if (il >= ir) break;

	SWAPI(ix[il],ix[ir])
      }
      SWAPI(ix[left+1],ix[ir]) // ir is the correct place for the pivot

      // the pivot belongs to the shortest interval:
      Nsmall = ir-begin; // [begin,ir)
      Nlarge = end-ir-1; // [ir+1,end)
      if (Nsmall > Nlarge) {
	Nlarge = Nsmall;
	*im = right = ir;
      }
      else
	/* note: if Nsmall == Nlarge (can only happen if right-left is
	   odd), then the right array will be shorter than the left by
	   one */
	*im = left = ir+1;
    }
    else {
      // code above does not work properly for less than three elements
      if (z[ix[left]] > z[ix[right-1]])
	SWAPI(ix[left],ix[right-1])
      *im = right-1;
      *z0 = z[ix[right-1]];
      break;
      /*** prefer a direct sort when right-left <= 7 or so? */ 
    }
  }

  // determine min and max
  if (min != NULL) {
    *min = *z0;
    for (int i = begin; i < *im; i++)
      if (z[ix[i]] < *min) *min = z[ix[i]];
  }
  if (max != NULL) {
    *max = *z0;
    for (int i = *im; i < end; i++)
      if (*max < z[ix[i]]) *max = z[ix[i]];
  }

  // re-adjust to get a better connectivity balance (recall that the
  // theta criterion is a geometric criterion)
  if (eta < 1.0) {
    const double z1_ = eta**z0+(1-eta)*z0_; // new 'balanced' midpoint

    // determine boundaries using the result from the first
    // (geometric) split
    if (*z0 <= z0_) {
      il = (left = *im)-1;
      ir = right = im0;
    }
    else {
      il = (left = im0)-1;
      ir = right = *im;
    }

    // as in MPexp_split_() below
    do il++; while (il < right && z[ix[il]] < z1_);
    do ir--; while (left <= ir && z1_ < z[ix[ir]]);
    if (il < ir) {
    SWAPI(ix[il],ix[ir])
      for ( ; ; ) {
	do il++; while (z[ix[il]] < z1_);
	do ir--; while (z1_ < z[ix[ir]]);
	if (il >= ir) break;
	SWAPI(ix[il],ix[ir])
      }
    }
    *im = il;
    *z0 = z1_;
  }
}
/*------------------------------------------------------------------------*/
void MPexp_split_(int begin,int *im,int end,double z0,
		  double *min,double *max,
		  int *ix,const double *z)
/* Split as above, but the coordinate z0 around which to split is
   already known. It is also assumed that min/max, whenever nonzero,
   are set to suitable initial guesses of the true min/max (typically
   the ones you get from MPexp_partition_() above). */
{
  int il = begin-1,ir = end;
  do il++; while (il < end && z[ix[il]] < z0);
  do ir--; while (begin <= ir && z0 < z[ix[ir]]);
  if (il < ir) {
    SWAPI(ix[il],ix[ir])
    for ( ; ; ) {
      do il++; while (z[ix[il]] < z0);
      do ir--; while (z0 < z[ix[ir]]);
      if (il >= ir) break;
      SWAPI(ix[il],ix[ir])
    }
  }
  *im = il;

  if (min != NULL) {
    for (int i = begin; i < *im; i++)
      if (z[ix[i]] < *min) *min = z[ix[i]];
  }
  if (max != NULL) {
    for (int i = *im; i < end; i++)
      if (*max < z[ix[i]]) *max = z[ix[i]];
  }
}
/*------------------------------------------------------------------------*/
void MPexp_sort_(MPexp *this,
		 int N,const double *zr,const double *zi,
		 int NE,const double *er,const double *ei,
		 int *maxm2p,int *levm2p,int Ndirect,double eta)
/* Sorts the N potentials (zr,zi) in the multipole tree this which is
   simultaneously constructed and, for NE > 0, also assigns each
   evaluation point (er,ei) a multipole box. After each level has been
   thus constructed the connectivity information is also
   determined. The integer maxm2p is set to the maximum number of
   cluster to cluster interactions in the whole tree (the maximum
   number of -1's in any column of all the connectivity matrices) and
   levm2p is the first level where a cluster to cluster interaction
   occurs. This function is critical to performance. */
{
  const int Nf = this->lptr[this->nlevel+1]-this->lptr[this->nlevel];
  int *ix,*ixptr,*ixptr0,*ixptr1;
  int *jx,*jxptr,*jxptr0,*jxptr1;
  mpSparse *C = this->connect;

  // (ixptr,ix) to be built
  ix = this->ix;
  for (int i = 0; i < N; i++) ix[i] = i;
  ixptr = this->ixptr;
  ixptr0 = &ixptr[Nf-1];
  ixptr0[0] = 0;
  ixptr0[1] = N;

  // ditto (jxptr,jx)
  if (NE) {
    jx = this->jx;
    for (int i = 0; i < NE; i++) jx[i] = i;
    jxptr = this->jxptr;
    jxptr0 = &jxptr[Nf-1];
    jxptr0[0] = 0;
    jxptr0[1] = NE;
  }
  else
    jx = jxptr = NULL;

  // root box
  this->root[0].z0 = 0.5*(this->xmin+this->xmax)+I*0.5*(this->ymin+this->ymax);
  this->root[0].d0 = 0.5*(this->xmax-this->xmin)+I*0.5*(this->ymax-this->ymin);

  /* connectivity: each matrix is strictly lower triangular since a
     box is always in near connection with itself and since
     connectivity is a symmetric relation */
  C[0].jcptr = mxMalloc(5*sizeof(int));
  C[0].jcptr[0] = 0;
  C[0].jcptr[1] = 0;
  // convenient in fmm2dGetMesh() above:
  C[0].kcptr = &C[0].jcptr[2];
  C[0].kcptr[0] = 0;
  C[0].kcptr[1] = 0; // note: 0 could be the final level!
  C[0].ir = &C[0].kcptr[2];
  C[0].ir[0] = 0;

  // loop over levels
  *maxm2p = 0;
  *levm2p = this->nlevel+1;
  for (int l = 1,Nmax = N,Nmax0,nc = 0; l <= this->nlevel; l++) {
    mpexp *parent = &this->root[this->lptr[l-1]];
    mpexp *child = &this->root[this->lptr[l]];
    const int N0 =  this->lptr[l]-this->lptr[l-1];
    const int N1 = this->lptr[l+1]-this->lptr[l];

    /* Step #1: divide each box on level l into 4 children such that
       about 1/4th of the potentials are located in each child box. */

    /* expanding arrays; they are written at in the lower part of the
       _same_ memory area (parents use the "0-part" and children the
       "1-part") */
    ixptr1 = &ixptr[Nf-N1];
    if (NE) jxptr1 = &jxptr[Nf-N1];

    /* determine the maximum number of coordinates after each split
       (the map is x --> ceil(0.5*x)) */
    if ((Nmax0 = (Nmax >> 1)+(Nmax&1)) <= Ndirect)
      Nmax0 = Nmax = Ndirect;
    else if ((Nmax = (Nmax0 >> 1)+(Nmax0&1)) <= Ndirect)
      Nmax = Ndirect;
    /* the intent with using Ndirect as a cutoff is to save some work
       at the final level whenever the number of particles is
       already sufficiently low */

    // loop over all parent boxes
    for (int i = 0,j = 0; i < N0; i++,j += 4) {
      dcmplx z,d;
      dcmplx z0,d0,z1,d1;
      dcmplx z00,d00,z01,d01,z10,d10,z11,d11;
      double split,min,max;

      // box to be split
      z = parent[i].z0;
      d = parent[i].d0;

      /* partition the coordinates [ixptr0[i],ixptr[i+1]) in box i
	 into 4 sections indexed by [ixptr1[4*i+n],ixptr1[4*i+n+1])
	 for n = 0..3 */
      ixptr1[4*i] = ixptr0[i];
      if (NE) jxptr1[4*i] = jxptr0[i];

      /* cut across the longest side if it is a certain factor longer,
	 otherwise take a y-cut followed by two x-cuts; this is the
	 Lebesgue ordering */
      if (creal(d) > meshratio*cimag(d)) {
	split = creal(z);
	MPexp_partition_(ixptr0[i],&ixptr1[4*i+2],ixptr0[i+1],&split,
			 &min,&max,
			 ix,zr,Nmax0,eta);
	if (NE)
	  MPexp_split_(jxptr0[i],&jxptr1[4*i+2],jxptr0[i+1],split,
		       &min,&max,jx,er);
#ifdef FMM_SHRINK
	z0 = 0.5*(min+split)+I*cimag(z);
	d0 = z0-(min+I*cimag(z-d));
	z1 = 0.5*(split+max)+I*cimag(z);
	d1 = (max+I*cimag(z+d))-z1;
#else
	z0 = 0.5*(creal(z-d)+split)+I*cimag(z);
	d0 = z0-(z-d);
	z1 = 0.5*(creal(z+d)+split)+I*cimag(z);
	d1 = (z+d)-z1;
#endif
      }
      else {
	split = cimag(z);
	MPexp_partition_(ixptr0[i],&ixptr1[4*i+2],ixptr0[i+1],&split,
			 &min,&max,
			 ix,zi,Nmax0,eta);
	if (NE)
	  MPexp_split_(jxptr0[i],&jxptr1[4*i+2],jxptr0[i+1],split,
		       &min,&max,jx,ei);
#ifdef FMM_SHRINK
	z0 = creal(z)+I*0.5*(min+split);
	d0 = z0-(creal(z-d)+I*min);
	z1 = creal(z)+I*0.5*(split+max);
	d1 = (creal(z+d)+I*max)-z1;
#else
	z0 = creal(z)+I*0.5*(cimag(z-d)+split);
	d0 = z0-(z-d);
	z1 = creal(z)+I*0.5*(cimag(z+d)+split);
	d1 = (z+d)-z1;
#endif
      }

      // same procedure for the resulting two boxes...
      if (meshratio*creal(d0) > cimag(d0)) {
	split = creal(z0);
	MPexp_partition_(ixptr0[i],&ixptr1[4*i+1],ixptr1[4*i+2],&split,
			 &min,&max,
			 ix,zr,Nmax,eta);
	if (NE)
	  MPexp_split_(jxptr0[i],&jxptr1[4*i+1],jxptr1[4*i+2],split,
		       &min,&max,jx,er);
#ifdef FMM_SHRINK
	z00 = 0.5*(min+split)+I*cimag(z0);
	d00 = z00-(min+I*cimag(z0-d0));
	z01 = 0.5*(split+max)+I*cimag(z0);
	d01 = (max+I*cimag(z0+d0))-z01;
#else
	z00 = 0.5*(creal(z0-d0)+split)+I*cimag(z0);
	d00 = z00-(z0-d0);
	z01 = 0.5*(creal(z0+d0)+split)+I*cimag(z0);
	d01 = (z0+d0)-z01;
#endif
      }
      else {
	split = cimag(z0);
	MPexp_partition_(ixptr0[i],&ixptr1[4*i+1],ixptr1[4*i+2],&split,
			 &min,&max,
			 ix,zi,Nmax,eta);
	if (NE)
	  MPexp_split_(jxptr0[i],&jxptr1[4*i+1],jxptr1[4*i+2],split,
		       &min,&max,jx,ei);
#ifdef FMM_SHRINK
	z00 = creal(z0)+I*0.5*(min+split);
	d00 = z00-(creal(z0-d0)+I*min);
	z01 = creal(z0)+I*0.5*(split+max);
	d01 = (creal(z0+d0)+I*max)-z01;
#else
	z00 = creal(z0)+I*0.5*(cimag(z0-d0)+split);
	d00 = z00-(z0-d0);
	z01 = creal(z0)+I*0.5*(cimag(z0+d0)+split);
	d01 = (z0+d0)-z01;
#endif
      }
      // and again...
      if (meshratio*creal(d1) > cimag(d1)) {
	split = creal(z1);
	MPexp_partition_(ixptr1[4*i+2],&ixptr1[4*i+3],ixptr0[i+1],&split,
			 &min,&max,
			 ix,zr,Nmax,eta);
	if (NE)
	  MPexp_split_(jxptr1[4*i+2],&jxptr1[4*i+3],jxptr0[i+1],split,
		       &min,&max,jx,er);
#ifdef FMM_SHRINK
	z10 = 0.5*(min+split)+I*cimag(z1);
	d10 = z10-(min+I*cimag(z1-d1));
	z11 = 0.5*(split+max)+I*cimag(z1);
	d11 = (max+I*cimag(z1+d1))-z11;
#else
	z10 = 0.5*(creal(z1-d1)+split)+I*cimag(z1);
	d10 = z10-(z1-d1);
	z11 = 0.5*(creal(z1+d1)+split)+I*cimag(z1);
	d11 = (z1+d1)-z11;
#endif
      }
      else {
	split = cimag(z1);
	MPexp_partition_(ixptr1[4*i+2],&ixptr1[4*i+3],ixptr0[i+1],&split,
			 &min,&max,
			 ix,zi,Nmax,eta);
	if (NE)
	  MPexp_split_(jxptr1[4*i+2],&jxptr1[4*i+3],jxptr0[i+1],split,
		       &min,&max,jx,ei);
#ifdef FMM_SHRINK
	z10 = creal(z1)+I*0.5*(min+split);
	d10 = z10-(creal(z1-d1)+I*min);
	z11 = creal(z1)+I*0.5*(split+max);
	d11 = (creal(z1+d1)+I*max)-z11;	
#else
	z10 = creal(z1)+I*0.5*(cimag(z1-d1)+split);
	d10 = z10-(z1-d1);
	z11 = creal(z1)+I*0.5*(cimag(z1+d1)+split);
	d11 = (z1+d1)-z11;
#endif
      }

      // produced boxes
      child[j].z0 = z00;
      child[j].d0 = d00;
      child[j+1].z0 = z01;
      child[j+1].d0 = d01;
      child[j+2].z0 = z10;
      child[j+2].d0 = d10;
      child[j+3].z0 = z11;
      child[j+3].d0 = d11;
    }

    // arrays grow to the left
    ixptr0 = ixptr1;
    if (NE) jxptr0 = jxptr1;

    /* Step #2: once all boxes at a level have been determined on,
       their connectivity can be sorted out. */

    /* allocate connectivity matrix using that all near connections to
       higher index boxes at the previous level implies 4 connections
       for each of 4 children at this level (and 6 additional
       connections to higher indices {0->[1 2 3],1->[2 3],2->[3]}
       between the children of each parent) */
    C[l].jcptr = mxMalloc(((2+(l == this->nlevel))*N1+1+
			   (6*N0+4*4*nc))*sizeof(int));
    C[l].jcptr[0] = 0;
    C[l].kcptr = &C[l].jcptr[N1+1];
    C[l].ir = &C[l].kcptr[(1+(l == this->nlevel))*N1];

    // setup connections
    nc = 0;  // counter for near connections
    for (int i = 0,j = 0; i < N0; i++)  // parent i
      for (int c = 0; c < 4; c++,j++) { // children j
	// forward- and backward counters in each column
	int fwd = C[l].jcptr[j];
	int bwd = fwd+3-c+4*(C[l-1].kcptr[i]-C[l-1].jcptr[i]);
	C[l].jcptr[j+1] = bwd;

	/* [@Lthetanoteplots] set up near connections to siblings > j
	   without checking the theta-criterion; the connection could
	   in principle be of the far type (except to the neighbour
	   sibling), but this should occur only very rarely */
	for (int d = 1; d < 4-c; d++)
	  C[l].ir[fwd++] = j+d;

	// parents (near) connections
	for (int ii = C[l-1].jcptr[i]; ii < C[l-1].kcptr[i]; ii++) {
	  /* the children of the near connections of box i become
	     either near connections or far connections of box j */
	  const int bwd0 = bwd;
	  const int K = C[l-1].ir[ii];

	  for (int d = 0, k = K << 2; d < 4; d++,k++)
	    // connection according to theta criterion
	    if (!mpexp_theta(&child[j],&child[k]))
	      // near connection
	      C[l].ir[fwd++] = k;
	    else
	      // far connection
	      C[l].ir[--bwd] = k;

	  /* if all 4 siblings are far connections, can they be
	     replaced with a single far connection to their parent
	     instead? */
	  if (bwd0-bwd == 4 && mpexp_theta(&child[j],&parent[K]))
	    C[l].ir[bwd] = K-N0; // N0 = child-parent
	  /* note: still uses 4 indices, the first of which is
	     negative and points to the parent K */
	}
	C[l].kcptr[j] = bwd;
	/* this splits the column into near connections
	   [jcptr[j],kcptr[j]) and far connections
	   [kcptr[j],jcptr[j+1]) */
	nc += bwd-C[l].jcptr[j];
	if (C[l].jcptr[j+1]-bwd > *maxm2p)
	  *maxm2p = C[l].jcptr[j+1]-bwd;
	if (l < *levm2p && *maxm2p > 0)
	  *levm2p = l;
 
	// last level: sort the near connections one step further
	if (l == this->nlevel) {
	  int dir1,dir2; // 1: cluster to particle, -1: particle to cluster
	  fwd = C[l].jcptr[j]-1;
	  bwd = C[l].kcptr[j];
	  for ( ; ; ) {
	    do
	      fwd++;
	    while (fwd < bwd && 
		   !(dir1 = mpexp_theta2(&child[j],&child[C[l].ir[fwd]])));
	    do {
	      bwd--;
	      if (fwd <= bwd && 
		  (dir2 = mpexp_theta2(&child[j],&child[C[l].ir[bwd]])))
		C[l].ir[bwd] *= dir2;
	      else
		break;
	    } while (1);

	    if (fwd >= bwd) break;
	    SWAPI(C[l].ir[fwd],C[l].ir[bwd])
            C[l].ir[bwd] *= dir1;
	  }
	  C[l].kcptr[j+N1] = bwd+1;
	  /* this splits the near connections into 'truly near'
	     [jcptr[j],kcptr[j+N]) and 'less near'
	     [kcptr[j+N],kcptr[j]) connections */
	}
      }
  }
}
/*------------------------------------------------------------------------*/
void MPexp_smooth_(int pot,SMOOTHER smooth,double xopt,bool cont,
		   double *cutoff,double *shape,double *scale)
/* Precomputes the three parameters (cutoff,shape,scale) given
   (pot,smooth,xopt,cutoff,cont). This yields a slicker code for
   mpexp_directInteract(). */
{
  *scale = 1.0;

  if (pot == 0)
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
      *cutoff = xopt*xopt;
      *shape = 0.5*log(*cutoff)-0.5;
#else
      *cutoff = xopt;
      *shape = log(*cutoff)-0.5;
#endif
      break;

    case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
      *cutoff = *cutoff**cutoff;
      *shape = xopt*xopt;
      if (cont)
	*scale = log(*cutoff)/log(*shape+*cutoff);
#else
      *shape = xopt*xopt;
      if (cont)
	*scale = 2.0*log(*cutoff)/log(*shape+*cutoff**cutoff);
#endif
      break;

    case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
    *cutoff = *cutoff**cutoff;
    *shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      *scale = log(*cutoff)/(log(*cutoff)+expint(*shape**cutoff));
#else
    *shape = 1.2564312086261696770/(xopt*xopt);
    if (cont)
      *scale = log(*cutoff)/(log(*cutoff)+0.5*expint(*shape**cutoff**cutoff));
#endif
      break;
    }
  else
    switch (smooth) {
    case DIRAC:
      break;

    case RANKINE:
#ifdef FMM_INLINECOMPLEX_DIRECT
      *cutoff = xopt*xopt;
      *shape = 1.0/(*cutoff);
#else
      *cutoff = xopt;
      *shape = 1.0/(*cutoff**cutoff);
#endif
      break;

    case SCULLY:
#ifdef FMM_INLINECOMPLEX_DIRECT
      *cutoff = *cutoff**cutoff;
      *shape = xopt*xopt;
      if (cont)
	*scale = 1.0+*shape/(*cutoff);
#else
      *shape = xopt*xopt;
      if (cont)
	*scale = 1.0+*shape/(*cutoff**cutoff);
#endif
      break;

    case OSEEN:
#ifdef FMM_INLINECOMPLEX_DIRECT
      *cutoff = *cutoff**cutoff;
      *shape = 1.2564312086261696770/(xopt*xopt);
      if (cont)
	*scale = -1.0/expm1(-*shape**cutoff);
#else
      *shape = 1.2564312086261696770/(xopt*xopt);
      if (cont)
	*scale = -1.0/expm1(-*shape**cutoff**cutoff);
#endif
      break;
    }
}
/*------------------------------------------------------------------------*/
void MPexp_eval2(const MPexp *this,
		 double *pr,double *pi,
		 const double *zr,const double *zi,
		 const double *mr,const double *mi,
		 int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the potentials themselves. Function
   is critical to performance. */
{
  // hopefully, some optimizing compiler can use this:
  const int Nf = this->lptr[this->nlevel+1]-this->lptr[this->nlevel];
  const mpexp *restrict offset = &this->root[this->lptr[this->nlevel]];
  const int *restrict ix = this->ix,*restrict ixptr = this->ixptr;
  const mpSparse *C = &this->connect[this->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  double shape,scale;

  MPexp_smooth_(pot,smooth,xopt,cont,&cutoff,&shape,&scale);
  for (int j = 0,jj; j < Nf; j++) {
    // far-field: contribution from distant potentials
    mpexp_eval(&offset[j],this->pcoeff,pr,pi,zr,zi,ix,ixptr[j],ixptr[j+1]);

    // near-field: symmetric interaction within list of near connections
    for (jj = jcptr[j]; jj < j2cptr[j]; jj++) {
      const int k = ir[jj];
      mpexp_directInteract2(pr,pi,zr,zi,mr,mi,ix,
			    ixptr[j],ixptr[j+1],
			    ixptr[k],ixptr[k+1],
			    pot,smooth,cutoff,shape,scale);
    }
    // cluster to particle interactions
    for ( ; jj < kcptr[j]; jj++) {
      const int k = ir[jj];
      if (k > 0)
	// multipole expansion in box j evaluated at all points in box k
	mpexp_evalmp(&offset[j],this->pcoeff,pot,pr,pi,zr,zi,
		     ix,ixptr[k],ixptr[k+1]);
      else
	mpexp_evalmp(&offset[-k],this->pcoeff,pot,pr,pi,zr,zi,
		     ix,ixptr[j],ixptr[j+1]);
    }

    // near-field: direct interaction within box
    mpexp_directInteract2(pr,pi,zr,zi,mr,mi,ix,ixptr[j],ixptr[j+1],
			  -1,ixptr[j+1],pot,smooth,cutoff,shape,scale);
  }
}
/*------------------------------------------------------------------------*/
void MPexp_eval(const MPexp *this,
		double *qr,double *qi,
		const double *er,const double *ei,
		const double *zr,const double *zi,
		const double *mr,const double *mi,
		int pot,SMOOTHER smooth,double xopt,double cutoff,bool cont)
/* Evaluation of the potential at the points (er,ei). Function is
   critical to performance. */
{
  const int Nf = this->lptr[this->nlevel+1]-this->lptr[this->nlevel];
  const mpexp *restrict offset = &this->root[this->lptr[this->nlevel]];
  const int *restrict ix = this->ix,*restrict ixptr = this->ixptr;
  const int *restrict jx = this->jx,*restrict jxptr = this->jxptr;
  const mpSparse *C = &this->connect[this->nlevel];
  const int *restrict jcptr = C->jcptr,*restrict j2cptr = &C->kcptr[Nf],
    *restrict kcptr = C->kcptr,*restrict ir = C->ir;
  double shape,scale;

  MPexp_smooth_(pot,smooth,xopt,cont,&cutoff,&shape,&scale);
  for (int j = 0,jj; j < Nf; j++) {
    // far-field: contribution from distant potentials
    mpexp_eval(&offset[j],this->pcoeff,qr,qi,er,ei,jx,jxptr[j],jxptr[j+1]);

    // near-field: symmetric interaction within list of near connections
    for (jj = jcptr[j]; jj < j2cptr[j]; jj++) {
      const int k = ir[jj];
      mpexp_directInteract(qr,qi,er,ei,
			   jx,jxptr[j],jxptr[j+1],
			   zr,zi,mr,mi,
			   ix,ixptr[k],ixptr[k+1],
			   pot,smooth,cutoff,shape,scale);
      mpexp_directInteract(qr,qi,er,ei,
			   jx,jxptr[k],jxptr[k+1],
			   zr,zi,mr,mi,
			   ix,ixptr[j],ixptr[j+1],
			   pot,smooth,cutoff,shape,scale);
    }
    // cluster to particle interactions
    for ( ; jj < kcptr[j]; jj++) {
      const int k = ir[jj];
      if (k > 0)
	// multipole expansion in box j evaluated in all points in box k
	mpexp_evalmp(&offset[j],this->pcoeff,pot,qr,qi,er,ei,
		     jx,jxptr[k],jxptr[k+1]);
      else
	mpexp_evalmp(&offset[-k],this->pcoeff,pot,qr,qi,er,ei,
		     jx,jxptr[j],jxptr[j+1]);
    }

    // near-field: direct interaction within box
    mpexp_directInteract(qr,qi,er,ei,
			 jx,jxptr[j],jxptr[j+1],
			 zr,zi,mr,mi,
			 ix,ixptr[j],ixptr[j+1],
			 pot,smooth,cutoff,shape,scale);
  }
}
/*------------------------------------------------------------------------*/
void mpexp_init(mpexp *this,int p,
                const double zr,const double zi,
		const double mr,const double mi,int pot)
/* Initializes a multipole expansion at the lowest level with a single
   pointmass (mr,mi) at (zr,zi). */
{
  /* We have that the potential is given by (pot = 0)
     mp(z) = -m*log(z-zi)
           = -m*log(z-z0-(zi-z0))
	   = -m*log((z-z0)*(1-(zi-z0)/(z-z0)))
	   = -m*[log(z-z0)+log(1-(zi-z0)/(z-z0))]
	   = -m*[log(z-z0)-sum_{j >= 1} (zi-z0)^j/j*(z-z0)^{-j},
     or othwerwise by (pot = 1)
     mp(z) = -m*1/(z-zi)
           = -m*1/(z-z0-(zi-z0))
	   = -m/(z-z0)*1/(1-(zi-z0)/(z-z0))
	   = -m*sum_{j >= 1} (zi-z0)^{j-1}*(z-z0)^{-j}.
  */
#ifndef FMM_INLINECOMPLEX
  const dcmplx r = (zr+I*zi)-this->z0;
  const dcmplx m = mr+I*mi;
  dcmplx *coeff = this->coeff1;

  if (pot == 0) {
    dcmplx z = r*m;
    coeff[0] -= m;
    for (int j = 1; j <= p; j++,z *= r) coeff[j] += z/j;
  }
  else {
    dcmplx z = m;
    for (int j = 1; j <= p; j++,z *= r) coeff[j] -= z;
  }
#else
  const double re = zr-creal(this->z0),im = zi-cimag(this->z0);
  dcmplx *coeff = this->coeff1;

  if (pot == 0) {
    double zre = mr,zim = mi;
    coeff[0] -= (mr+I*mi);
    for (int j = 1; j <= p; j++) {
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
      coeff[j] += (zre/j+I*zim/j);
    }
  }
  else {
    double zre = mr,zim = mi;
    for (int j = 1; j <= p; j++) {
      coeff[j] -= (zre+I*zim);
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
    }
  }
#endif
}
/*------------------------------------------------------------------------*/
void mpexp_initp(mpexp *this,int p,
		 const double zr,const double zi,
		 const double mr,const double mi,int pot)
/* Initializes a multipole expansion at the lowest level with a single
   pointmass (mr,mi) at (zr,zi). The pointmass is assumed to be
   sufficiently far away from the center of the box that a polynomial
   expansion may be used directly. */
{
  /* We have that the potential is given by (pot = 0)
     mp(z) = -m*log(z-zi)
           = -m*log(z0-zi+(z-z0))
	   = -m*[log(z0-zi)-sum_{j >= 1} (z-z0)^j/j*(zi-z0)^{-j},
     or othwerwise by (pot = 1)
     mp(z) = -m*1/(z-zi)
           = -m*1/(z0-zi+(z-z0))
	   = m*sum_{j >= 1} (z-z0)^{j-1}*(zi-z0)^{-j}.
  */
#ifndef FMM_INLINECOMPLEX
  const dcmplx r = 1.0/((zr+I*zi)-this->z0);
  const dcmplx m = mr+I*mi;
  dcmplx *coeff = this->coeff2;

  if (pot == 0) {
    dcmplx z = m*r;
    coeff[0] += m*clog(-r);
    for (int j = 1; j <= p; j++,z *= r) coeff[j] += z/j;
  }
  else {
    dcmplx z = m*r;
    for (int j = 0; j <= p; j++,z *= r) coeff[j] += z;
  }
#else
  const dcmplx r = 1.0/((zr+I*zi)-this->z0);
  const double re = creal(r),im = cimag(r);
  dcmplx *coeff = this->coeff2;

  if (pot == 0) {
    double zre = mr*re-mi*im,zim = mi*re+mr*im;
    {
      const dcmplx logr = clog(-r);
      coeff[0] += (mr*creal(logr)-mi*cimag(logr)+
		   I*(mi*creal(logr)+mr*cimag(logr)));
    }
    for (int j = 1; j <= p; j++) {
      coeff[j] += (zre/j+I*zim/j);
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
    }
  }
  else {
    double zre = mr*re-mi*im,zim = mi*re+mr*im;
    for (int j = 0; j <= p; j++) {
      coeff[j] += (zre+I*zim);
      double zre0 = zre*re-zim*im;
      zim = zim*re+zre*im;
      zre = zre0;
    }
  }
#endif
}
/*------------------------------------------------------------------------*/
bool mpexp_theta(const mpexp *this,const mpexp *that)
/* Theta criterion for accurate cluster to cluster interaction between
   two boxes this and that.

   This routine is a workhorse and needs to be fast.

   The maximum relative error in using a multipole expansion from a
   box 'this' to evaluate the field in a box 'that' is error <=
   constant * (R/|R+minimum distance between the boxes|)^(p+1) for
   points |z| > R. Using that the minimum distance is >= d-R-r (where
   d is the center-to-center distance and where r is the radius of the
   smaller box), we are led to the requirement that R/(d-r) <= theta
   and this is what is implemented below (see [4] for the full
   analysis).

   This also means that theta = 1/2 is equivalent to the boxes being
   classically well-separated as in [1,2]. With a general theta, any
   of the two sets may be expanded by a factor of 1/theta and
   arbitrarily rotated about its center point without touching the
   other set.

   However, in that case one should note that the theta used in the
   classical uniform multipole method [2] is actually 1/(2*sqrt(2)-1)
   = 0.5469181 and not 1/2 -- perhaps for convenience, given the
   quadratic boxes.

   Also, in [3] a different and less stringent criterion is used; R <=
   c*d. For uniform boxes, their c = 1/2 would mean theta = sqrt(2)/3
   = 0.4714045 here. Note that trivially, R <= theta/(1+theta)*d
   implies R+theta*r <= theta*d, but the former criterion is less
   sharp when the two boxes have different sizes.

   A slightly sharper result stems from squaring and using Young's
   weighted inequality; then R^2+theta*r^2 <= theta^2/(1+theta)*d^2
   implies R+theta*r <= theta*d. This is used below as a first check
   in order to avoid computing the square-root.

   Also, note that the GKZ-criterion in [3] yields an error
   proportional to (c/(1-2*c))^(p+1) and not to c^(p+1) as they
   claim. Hence c = theta/(1+2*theta) would be appropriate -- this is
   even less sharp compared to the trivial interpretation mentioned
   above. Unfortunately, following their method of proof, the
   currently implemented criterion yields an error proportional to
   (theta/(1-theta))^(p+1), which would suggest that the choice theta
   = 1/2 should be avoided.

   Notes: the GKZ-criterion is found in [3], Eq. (8.31) at p. 379 and
   the error estimate is derived at p. 386. The first estimate [3] in
   (8.40) is incorrect (change (1+2*theta) to (1-2*theta) and the
   second is unsharp (replace (1-2*theta) with 1/(1+2*theta)). In the
   proof that follows, the problematic factor (1-2*theta) is
   inadvertently omitted.

   References:
     [1] J. Carrier, L. Greengard, and V. Rokhlin: "A fast adaptive
         multipole algorithm for particle simulations". SIAM
         J. Sci. Stat. Comput., 9(4):669--686, 1988.
     [2] L. Greengard and V. Rokhlin: "A fast algorithm for particle
         simulations". J.Comput.Phys., 73:325--348, 1987.
     [3] M. Griebel, S. Knapek, and G. Zumbusch: "Numerical Simulation
         in Molecular Dynamics", volume 5 of Texts in Computational
         Science and Engineering. Springer Verlag, Berlin, 2007.
     [4] S. Engblom: "On well-separated sets and fast mutlipole
         methods", 2010. Available at http://arxiv.org/abs/1006.2269.
*/
{
  const double theta2_ = theta*theta;
  const double thetac_ = theta2_/(1.0+theta);
  // avoid cabs():
  const double thisrad = creal(this->d0)*creal(this->d0)+
    cimag(this->d0)*cimag(this->d0);
  const double thatrad = creal(that->d0)*creal(that->d0)+
    cimag(that->d0)*cimag(that->d0);
  const dcmplx d = this->z0-that->z0;
  const double d2 = creal(d)*creal(d)+cimag(d)*cimag(d);

  /* note: criterion uses strict inequality to get rid of the singular
     case when both boxes are the same points (this can happen!) */
  if (thisrad >= thatrad) {
    // sufficient condition (return probability is about 80%)
    if (thisrad+theta*thatrad < thetac_*d2)
      return true;
    else
      return thisrad+theta2_*thatrad+
	2.0*theta*sqrt(thisrad*thatrad) < theta2_*d2;
  }
  else {
    if (thatrad+theta*thisrad < thetac_*d2)
      return true;
    else
      return thatrad+theta2_*thisrad+
	2.0*theta*sqrt(thatrad*thisrad) < theta2_*d2;
  }
}
/*------------------------------------------------------------------------*/
int mpexp_theta2(const mpexp *this,const mpexp *that)
/* 2nd theta criterion: if mpexp_theta() above is false, is it due to
   only one of the boxes? Then the multipoles of the larger box can be
   shifted into a polynomial expansion in the smaller, and the
   multipole expansion in the smaller can be evaluated at each point
   in the larger box. Returns -1 if this is the smaller box (cluster
   to particle) and 1 otherwise (particle to cluster). */
{
  const double theta2_ = theta*theta;
  const double thetac_ = theta2_/(1.0+theta);
  const double thisrad = creal(this->d0)*creal(this->d0)+
    cimag(this->d0)*cimag(this->d0);
  const double thatrad = creal(that->d0)*creal(that->d0)+
    cimag(that->d0)*cimag(that->d0);
  const dcmplx d = this->z0-that->z0;
  const double d2 = creal(d)*creal(d)+cimag(d)*cimag(d);

  // derived as a kind of adjoint to mpexp_theta() above
  if (thatrad >= thisrad) {
    if (thisrad+theta*thatrad < thetac_*d2)
      return 1;
    else
      return thisrad+theta2_*thatrad+
	2.0*theta*sqrt(thisrad*thatrad) < theta2_*d2;
  }
  else {
    if (thatrad+theta*thisrad < thetac_*d2)
      return -1;
    else
      return -(thatrad+theta2_*thisrad+
	       2.0*theta*sqrt(thatrad*thisrad) < theta2_*d2);
  }
}
/*------------------------------------------------------------------------*/
void mpexp_eval(const mpexp *this,int p,
		double *vr,double *vi,
		const double *zr,const double *zi,
		const int *ix,int begin,int end)
/* Evaluation of near field at coordinates (zr,zi). */
{
  const dcmplx *coeff = this->coeff2;

  if (vi != NULL)
    for (int i = begin; i < end; i++) {
#ifndef FMM_INLINECOMPLEX
      dcmplx polyval = coeff[p];
      const dcmplx r = (zr[ix[i]]+I*zi[ix[i]])-this->z0;

      // Horner's method
      for (int j = p-1; j >= 0; j--) polyval = r*polyval+coeff[j];

      // add the result
      vr[ix[i]] += creal(polyval);
      vi[ix[i]] += cimag(polyval);
#else
      const double re = zr[ix[i]]-creal(this->z0),im = zi[ix[i]]-cimag(this->z0);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);
      
      // Horner's method
      for (int j = p-1; j >= 0; j--) {
	double pre0;
	pre0 = re*pre-im*pim+creal(coeff[j]);
	pim = re*pim+im*pre+cimag(coeff[j]);
	pre = pre0;
      }

      // add the result
      vr[ix[i]] += pre;
      vi[ix[i]] += pim;
#endif
    }
  else
    for (int i = begin; i < end; i++) {
#ifndef FMM_INLINECOMPLEX
      dcmplx polyval = coeff[p];
      const dcmplx r = (zr[ix[i]]+I*zi[ix[i]])-this->z0;
      for (int j = p-1; j >= 0; j--) polyval = r*polyval+coeff[j];
      vr[ix[i]] += creal(polyval);
#else
      const double re = zr[ix[i]]-creal(this->z0),im = zi[ix[i]]-cimag(this->z0);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);
      for (int j = p-1; j >= 0; j--) {
	double pre0;
	pre0 = re*pre-im*pim+creal(coeff[j]);
	pim = re*pim+im*pre+cimag(coeff[j]);
	pre = pre0;
      }
      vr[ix[i]] += pre;
#endif
    }
}
/*------------------------------------------------------------------------*/
void mpexp_evalmp(const mpexp *this,int p,int pot,
		  double *vr,double *vi,
		  const double *zr,const double *zi,
		  const int *ix,int begin,int end)
/* Evaluation of multipole expansion at coordinates (zr,zi). */
{
  const dcmplx *coeff = this->coeff1;

  if (vi != NULL) {
    for (int i = begin; i < end; i++) {
#ifndef FMM_INLINECOMPLEX
      dcmplx polyval = coeff[p];
      const dcmplx r = 1.0/((zr[ix[i]]+I*zi[ix[i]])-this->z0);

      // Horner's method for inverse polynomial
      for (int j = p-1; j > 0; j--) polyval = r*polyval+coeff[j];
      if (pot == 0)
	polyval = r*polyval-coeff[0]*clog(r);
      else
	polyval *= r;

      vr[ix[i]] += creal(polyval);
      vi[ix[i]] += cimag(polyval);
#else
      const dcmplx r = 1.0/((zr[ix[i]]+I*zi[ix[i]])-this->z0);
      const double re = creal(r),im = cimag(r);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);

      // Horner's method for inverse polynomial
      for (int j = p-1; j > 0; j--) {
	double pre0;
	pre0 = re*pre-im*pim+creal(coeff[j]);
	pim = re*pim+im*pre+cimag(coeff[j]);
	pre = pre0;
      }
      {
	double pre0;
	pre0 = re*pre-im*pim;
	pim = re*pim+im*pre;
	pre = pre0;
	if (pot == 0) {
	  const dcmplx logr = clog(r);
	  pre -= creal(coeff[0])*creal(logr)-cimag(coeff[0])*cimag(logr);
	  pim -= creal(coeff[0])*cimag(logr)+creal(coeff[0])*cimag(logr);
	}
      }

      vr[ix[i]] += pre;
      vi[ix[i]] += pim;
#endif
    }
  }
  else {
    for (int i = begin; i < end; i++) {
#ifndef FMM_INLINECOMPLEX
      dcmplx polyval = coeff[p];
      const dcmplx r = 1.0/((zr[ix[i]]+I*zi[ix[i]])-this->z0);
      for (int j = p-1; j > 0; j--) polyval = r*polyval+coeff[j];
      if (pot == 0)
	polyval = r*polyval-coeff[0]*clog(r);
      else
	polyval *= r;
      vr[ix[i]] += creal(polyval);
#else
      const dcmplx r = 1.0/((zr[ix[i]]+I*zi[ix[i]])-this->z0);
      const double re = creal(r),im = cimag(r);
      double pre = creal(coeff[p]),pim = cimag(coeff[p]);

      // Horner's method for inverse polynomial
      for (int j = p-1; j > 0; j--) {
	double pre0;
	pre0 = re*pre-im*pim+creal(coeff[j]);
	pim = re*pim+im*pre+cimag(coeff[j]);
	pre = pre0;
      }
      {
	double pre0;
	pre0 = re*pre-im*pim;
	pim = re*pim+im*pre;
	pre = pre0;
	if (pot == 0) {
	  const dcmplx logr = clog(r);
	  pre -= creal(coeff[0])*creal(logr)-cimag(coeff[0])*cimag(logr);
	  pim -= creal(coeff[0])*cimag(logr)+creal(coeff[0])*cimag(logr);
	}
      }

      vr[ix[i]] += pre;
#endif
    }
  }
}
/*------------------------------------------------------------------------*/
#define BEGIN(begin,i) (begin == -1 ? i+1 : begin)

void mpexp_directInteract2(double *pr,double *pi,
			   const double *zr,const double *zi,
			   const double *mr,const double *mi,
			   const int *ix,
			   int begin1,int end1,int begin2,int end2,
			   int pot,SMOOTHER smooth,
			   double cutoff,double shape,double scale)
/* Direct interaction between potentials inside a box (begin2 == -1)
   or between two boxes. */
{
  if (pot == 0) { // logaritmic potential
    if (mi == NULL) { // rlog2()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double P = 0.5*log(re*re+im*im);
#else
	    const double P = log(hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]));
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1)
	    pr[ix[i]] -= mr[ix[i]]*shape;
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*(ad/cutoff)+shape;
#else
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	}
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1)
	    pr[ix[i]] -= mr[ix[i]]*0.5*log(shape)*scale;
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*log(shape+ad)*scale;
#else
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*log(shape+ad*ad)*scale;
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	}
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1)
	    pr[ix[i]] += mr[ix[i]]*0.5*(log(shape)+gamma_)*scale;
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else if (ad != 0.0)
	      P = 0.5*(log(ad)+expint(shape*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#else
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else if (ad != 0.0)
	      P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	  }
	}
	break;
      }
    }
    else { // zlog2()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double P = 0.5*log(re*re+im*im);
#else
	    const double P = log(hypot(zr[ix[i]]-zr[ix[j]],
				       zi[ix[i]]-zi[ix[j]]));
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1) {
	    pr[ix[i]] -= mr[ix[i]]*shape;
	    pi[ix[i]] -= mi[ix[i]]*shape;
	  }
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*(ad/cutoff)+shape;
#else
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	}
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1) {
	    pr[ix[i]] -= mr[ix[i]]*0.5*log(shape)*scale;
	    pi[ix[i]] -= mi[ix[i]]*0.5*log(shape)*scale;
	  }
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*log(shape+ad)*scale;
#else
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*log(shape+ad*ad)*scale;
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	}
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++) {
	  if (begin2 == -1) {
	    pr[ix[i]] += mr[ix[i]]*0.5*(log(shape)+gamma_)*scale;
	    pi[ix[i]] += mi[ix[i]]*0.5*(log(shape)+gamma_)*scale;
	  }
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else if (ad != 0.0)
	      P = 0.5*(log(ad)+expint(shape*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#else
	    const double ad = hypot(zr[ix[i]]-zr[ix[j]],zi[ix[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else if (ad != 0.0)
	      P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#endif
	    pr[ix[i]] -= mr[ix[j]]*P;
	    pi[ix[i]] -= mi[ix[j]]*P;
	    pr[ix[j]] -= mr[ix[i]]*P;
	    pi[ix[j]] -= mi[ix[i]]*P;
	  }
	}
	break;
      }
    }
  }
  else { // harmonic potential
    if (mi == NULL) { // rinv2()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double P = 1.0/(re*re+im*im);
	    pr[ix[i]] -= mr[ix[j]]*P*re;
	    pi[ix[i]] += mr[ix[j]]*P*im;
	    pr[ix[j]] += mr[ix[i]]*P*re;
	    pi[ix[j]] -= mr[ix[i]]*P*im;
#else
	    const dcmplx P = 1.0/(zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]));
	    pr[ix[i]] -= mr[ix[j]]*creal(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P);
#endif
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = shape;
	    pr[ix[i]] -= mr[ix[j]]*P*re;
	    pi[ix[i]] += mr[ix[j]]*P*im;
	    pr[ix[j]] += mr[ix[i]]*P*re;
	    pi[ix[j]] -= mr[ix[i]]*P*im;
#else
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)*shape;
	    pr[ix[i]] -= mr[ix[j]]*creal(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P);
#endif
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = 1.0/(shape+ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*P*re;
	    pi[ix[i]] += mr[ix[j]]*P*im;
	    pr[ix[j]] += mr[ix[i]]*P*re;
	    pi[ix[j]] -= mr[ix[i]]*P*im;
#else
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)/(shape+ad*ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*creal(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P);
#endif
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad != 0.0) {
	      if (ad >= cutoff)
		P = 1.0/ad;
	      else
		P = -expm1(-shape*ad)/ad*scale;
	      pr[ix[i]] -= mr[ix[j]]*P*re;
	      pi[ix[i]] += mr[ix[j]]*P*im;
	      pr[ix[j]] += mr[ix[i]]*P*re;
	      pi[ix[j]] -= mr[ix[i]]*P*im;
	    }
#else
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    if (ad != 0.0) {
	      dcmplx P;
	      if (ad >= cutoff)
		P = 1.0/d;
	      else
		P = -expm1(-shape*ad*ad)/d*scale;
	      pr[ix[i]] -= mr[ix[j]]*creal(P);
	      pi[ix[i]] -= mr[ix[j]]*cimag(P);
	      pr[ix[j]] += mr[ix[i]]*creal(P);
	      pi[ix[j]] += mr[ix[i]]*cimag(P);
	    }
#endif
	  }
	break;
      }
    }
    else { // zinv2()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double P = 1.0/(re*re+im*im);
	    pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	    pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
	    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
	    pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
#else
	    const dcmplx P = 1.0/(zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]));
	    pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
#endif
	  }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = shape;
	    pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	    pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
	    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
	    pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
#else
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)*shape;
	    pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
#endif
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = 1.0/(shape+ad)*scale;
	    pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	    pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
	    pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
	    pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
#else
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)/(shape+ad*ad)*scale;
	    pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	    pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
#endif
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = BEGIN(begin2,i); j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = zr[ix[i]]-zr[ix[j]],im = zi[ix[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad != 0.0) {
	      if (ad >= cutoff)
		P = 1.0/ad;
	      else
		P = -expm1(-shape*ad)/ad*scale;
	      pr[ix[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	      pi[ix[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
	      pr[ix[j]] += (mr[ix[i]]*re+mi[ix[i]]*im)*P;
	      pi[ix[j]] -= (mr[ix[i]]*im-mi[ix[i]]*re)*P;
	    }
#else
	    const dcmplx d = zr[ix[i]]-zr[ix[j]]+I*(zi[ix[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    if (ad != 0.0) {
	      dcmplx P;
	      if (ad >= cutoff)
		P = 1.0/d;
	      else
		P = -expm1(-shape*ad*ad)/d*scale;
	      pr[ix[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	      pi[ix[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	      pr[ix[j]] += mr[ix[i]]*creal(P)-mi[ix[i]]*cimag(P);
	      pi[ix[j]] += mr[ix[i]]*cimag(P)+mi[ix[i]]*creal(P);
	    }
#endif
	  }
	break;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
void mpexp_directInteract(double *qr,double *qi,
			  const double *er,const double *ei,
			  const int *jx,int begin1,int end1,
			  const double *zr,const double *zi,
			  const double *mr,const double *mi,
			  const int *ix,int begin2,int end2,
			  int pot,SMOOTHER smooth,
			  double cutoff,double shape,double scale)
/* Direct evaluation of field from potentials inside one box computed
   at points (er,ei) inside another box. */
{
  if (pot == 0) { // logarithmic potential
    if (mi == NULL) { // rlog()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++)
	    if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	      const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	      const double P = 0.5*log(re*re+im*im);
#else
	      const double P = log(hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]));
#endif
	      qr[jx[i]] -= mr[ix[j]]*P;
	    }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*(ad/cutoff)+shape;
#else
	    const double ad = hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	    qr[jx[i]] -= mr[ix[j]]*P;
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*log(shape+ad)*scale;
#else
	    const double ad = hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*log(shape+ad*ad)*scale;
#endif
	    qr[jx[i]] -= mr[ix[j]]*P;
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else if (ad != 0.0)
	      P = 0.5*(log(ad)+expint(shape*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#else
	    const double ad = hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else if (ad != 0.0)
	      P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#endif
	    qr[jx[i]] -= mr[ix[j]]*P;
	  }
	break;
      }
    }
    else { // zlog()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++)
	    if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	      const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	      const double P = 0.5*log(re*re+im*im);
#else
	      const double P = log(hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]));
#endif
	      qr[jx[i]] -= mr[ix[j]]*P;
	      qi[jx[i]] -= mi[ix[j]]*P;
	    }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*(ad/cutoff)+shape;
#else
	    const double ad = hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*(ad/cutoff)*(ad/cutoff)+shape;
#endif
	    qr[jx[i]] -= mr[ix[j]]*P;
	    qi[jx[i]] -= mi[ix[j]]*P;
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else
	      P = 0.5*log(shape+ad)*scale;
#else
	    const double ad = hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else
	      P = 0.5*log(shape+ad*ad)*scale;
#endif
	    qr[jx[i]] -= mr[ix[j]]*P;
	    qi[jx[i]] -= mi[ix[j]]*P;
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 0.5*log(ad);
	    else if (ad != 0.0)
	      P = 0.5*(log(ad)+expint(shape*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#else
	    const double ad = hypot(er[jx[i]]-zr[ix[j]],ei[jx[i]]-zi[ix[j]]);
	    double P;
	    if (ad >= cutoff)
	      P = log(ad);
	    else if (ad != 0.0)
	      P = (log(ad)+0.5*expint(shape*ad*ad))*scale;
	    else
	      P = -0.5*(log(shape)+gamma_)*scale;
#endif
	    qr[jx[i]] -= mr[ix[j]]*P;
	    qi[jx[i]] -= mi[ix[j]]*P;
	  }
	break;
      }
    }
  }
  else { // harmonic potential
    if (mi == NULL) { // rinv()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++)
	    if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	      const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	      const double P = 1.0/(re*re+im*im);
	      qr[jx[i]] -= mr[ix[j]]*P*re;
	      qi[jx[i]] += mr[ix[j]]*P*im;
#else
	      const dcmplx P = 1.0/(er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]));
	      qr[jx[i]] -= mr[ix[j]]*creal(P);
	      qi[jx[i]] -= mr[ix[j]]*cimag(P);
#endif
	    }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = shape;
	    qr[jx[i]] -= mr[ix[j]]*P*re;
	    qi[jx[i]] += mr[ix[j]]*P*im;
#else
	    const dcmplx d = er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)*shape;
	    qr[jx[i]] -= mr[ix[j]]*creal(P);
	    qi[jx[i]] -= mr[ix[j]]*cimag(P);
#endif
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = 1.0/(shape+ad)*scale;
	    qr[jx[i]] -= mr[ix[j]]*P*re;
	    qi[jx[i]] += mr[ix[j]]*P*im;
#else
	    const dcmplx d = er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)/(shape+ad*ad)*scale;
	    qr[jx[i]] -= mr[ix[j]]*creal(P);
	    qi[jx[i]] -= mr[ix[j]]*cimag(P);
#endif
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad != 0.0) {
	      if (ad >= cutoff)
		P = 1.0/ad;
	      else
		P = -expm1(-shape*ad)/ad*scale;
	      qr[jx[i]] -= mr[ix[j]]*P*re;
	      qi[jx[i]] += mr[ix[j]]*P*im;
	    }
#else
	    const dcmplx d = er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    if (ad != 0.0) {
	      dcmplx P;
	      if (ad >= cutoff)
		P = 1.0/d;
	      else
		P = -expm1(-shape*ad*ad)/d*scale;
	      qr[jx[i]] -= mr[ix[j]]*creal(P);
	      qi[jx[i]] -= mr[ix[j]]*cimag(P);
	    }
#endif
	  }
	break;
      }
    }
    else { // zinv()
      switch (smooth) {
      case DIRAC:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++)
	    if (er[jx[i]] != zr[ix[j]] || ei[jx[i]] != zi[ix[j]]) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	      const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	      const double P = 1.0/(re*re+im*im);
	      qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	      qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
#else
	      const dcmplx P = 1.0/(er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]));
	      qr[jx[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	      qi[jx[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
#endif
	    }
	break;

      case RANKINE:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = shape;
	    qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	    qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
#else
	    const dcmplx d = er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)*shape;
	    qr[jx[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    qi[jx[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
#endif
	  }
	break;

      case SCULLY:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad >= cutoff)
	      P = 1.0/ad;
	    else
	      P = 1.0/(shape+ad)*scale;
	    qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	    qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
#else
	    const dcmplx d = er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    dcmplx P;
	    if (ad >= cutoff)
	      P = 1.0/d;
	    else
	      P = conj(d)/(shape+ad*ad)*scale;
	    qr[jx[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	    qi[jx[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
#endif
	  }
	break;

      case OSEEN:
	for (int i = begin1; i < end1; i++)
	  for (int j = begin2; j < end2; j++) {
#ifdef FMM_INLINECOMPLEX_DIRECT
	    const double re = er[jx[i]]-zr[ix[j]],im = ei[jx[i]]-zi[ix[j]];
	    const double ad = re*re+im*im;
	    double P;
	    if (ad != 0.0) {
	      if (ad >= cutoff)
		P = 1.0/ad;
	      else
		P = -expm1(-shape*ad)/ad*scale;
	      qr[jx[i]] -= (mr[ix[j]]*re+mi[ix[j]]*im)*P;
	      qi[jx[i]] += (mr[ix[j]]*im-mi[ix[j]]*re)*P;
	    }
#else
	    const dcmplx d = er[jx[i]]-zr[ix[j]]+I*(ei[jx[i]]-zi[ix[j]]);
	    const double ad = cabs(d);
	    if (ad != 0.0) {
	      dcmplx P;
	      if (ad >= cutoff)
		P = 1.0/d;
	      else
		P = -expm1(-shape*ad*ad)/d*scale;
	      qr[jx[i]] -= mr[ix[j]]*creal(P)-mi[ix[j]]*cimag(P);
	      qi[jx[i]] -= mr[ix[j]]*cimag(P)+mi[ix[j]]*creal(P);
	    }
#endif
	  }
	break;
      }
    }
  }
}
/*------------------------------------------------------------------------*/
