/* fmm.h -- Fast multipole method for 2-D potentials in free space */

/* S. Engblom 2009-10-08 (Revision) */
/* S. Engblom 2009-04-15 (Major revision, adaptivity) */
/* S. Engblom 2007-07-08 (Major revision, performance) */
/* S. Engblom 2007-06-19 */

#ifndef __fmm_h
#define __fmm_h

#include <math.h>
#include <complex.h>
#include <stdbool.h>

const static double gamma_ = 0.5772156649015328606065121;

typedef enum {SCULLY = 0,RANKINE = 1,OSEEN = 2,DIRAC = 3,
	      ERR_SMOOTHER = -1} SMOOTHER;

typedef double complex dcmplx;

// for performance experiments:
#ifdef PRINTTIME
#include <time.h>
static clock_t TIME_before,TIME_after;
#define StartTime TIME_before = clock()
#define StopTime TIME_after = clock()
#define PrintTime(s) mexPrintf(s, \
          (double)(TIME_after-TIME_before)/CLOCKS_PER_SEC)
#else
#define StartTime
#define StopTime
#define PrintTime(s)
#endif /* PRINTTIME */

#include "mex.h"

/* Criterion for near/far approximation: two boxes interact through
   cluster-cluster approximations provided that the function
   mpexp_theta() below returns true. Generally, small theta (say, less
   than 0.5) means that less coefficients are needed but that the
   interaction lists get larger, and vice versa. */
#ifndef FMM_THETA
#define FMM_THETA 0.5
#endif
const static double theta = FMM_THETA; // 0 < theta < 1

/* Criterion for dividing a box. If one side is more than meshratio
   times the other, then the longer side is split into two. Otherwise
   it is split according to an ordering which yields good cache
   effects. */
#ifndef FMM_MESHRATIO
#define FMM_MESHRATIO 2.0
#endif
const static double meshratio = FMM_MESHRATIO; // 1.0 <= meshratio

/* Multipole expansion structure. The coefficients are to be
   understood as
      mpexp(z) = coeff1[0]*log(z-z0)+
      coeff1[1]/(z-z0)+coeff1[2]/(z-z0)^2+coeff1[3]/(z-z0)^3+...
   or, depending on the stage of the algorithm,
      mpexp(z) = coeff2[0]+coeff2[1]*(z-z0)+coeff2[2]*(z-z0)^2+...
*/
typedef struct {
  dcmplx z0,d0;           // midpoint and radius (z0 to upper right corner)
  dcmplx *coeff1,*coeff2; // multipole coefficients
} mpexp;

/* Sparse connectivity matrix (special CCS format integer matrix): the
   row indices in ir of column j are found in [jcptr[j],jcptr[j+1]),
   the near connections (think of these as the value 1) in
   [jcptr[j],kcptr[j]) and the far connections (i.e. value -1) in
   [kcptr[j],jcptr[j+1]). The row indices of the near part is sorted
   in ascending- and the far part in descending order,
   respectively. Only the strictly lower triangular part is stored.

   For far connections between two levels (value -2), a negative index
   will occur in ir, pointing to a box at the previous levels. Then
   three positive indices follow, pointing to children of the previous
   negative index. Those indices are ignored in the FMM assembly and are
   a by-product of the algorithm.

   At the finest level a special syntax is used in which the near part
   is grouped into a 'truly' near part (value 1) in [jptr[j],kcptr[j+N])
   and a 'less strongly' connected part (value 2) in
   [kcptr[j+N],kcptr[j]). In this case the near part is not sorted and
   also, the rowindices are signed (positive for cluster to particle,
   negative for particle to cluster connection). */
typedef struct {
  int *jcptr,*kcptr; /* column pointer and pointer to the beginning of
			the 'far'-part */
  int *ir;           // row indices
} mpSparse;

/* Multipole expansion tree in vectorized form. First comes the root,
   then the 4 children of the root and so on */
typedef struct {
  double xmin,xmax,
    ymin,ymax;       // bounding box
  int pcoeff;        /* number of coefficients offset by one; coeff[0]
			through coeff[pcoeff] */
  int nlevel;        /* number of levels offset by one; level =
			0..nlevel */
  int *lptr;         /* pointer to levels in tree; level l can be
			found in root[i] where lptr[l] <= i <
			lptr[l+1] and lptr[nlevel+1] is the total
			number of boxes. */
  mpexp *root;       // box 0, the root of the tree
  mpSparse *connect; // connectivity information for each level
  int *ix,*ixptr;    // permutation of potentials
  int *jx,*jxptr;    // permutation of evaluation points
} MPexp;

/* Evaluation by the fast multipole method to accuracy TOL of the
   potential from N pointmasses (mr,mi) at the positions (zr,zi). */
void fmm2dInteract(int N,
		   const double *zr,const double *zi,
		   const double *mr,const double *mi,
		   int NE,
		   const double *er,const double *ei,
		   double *pr,double *pi,
		   double *qr,double *qi,
		   int pot,double tol,int Ndirect,double eta,
		   SMOOTHER smooth,double xopt,double cutoff,bool cont);

/* Returns the multipole mesh and sparse connection matrices. For
   debugging and pedagogical purposes. */
void fmm2dGetMesh(int nlhs,mxArray *plhs[],
		  int N,
		  const double *zr,const double *zi,
		  const double *mr,const double *mi,
		  int NE,
		  const double *er,const double *ei,
		  int pot,double tol,int Ndirect,double eta);

#endif /* __fmm_h */
