/* fmmshift.c */

/* S. Engblom 2010-02-18 */

#include "fmmshift.h"

// deallocate persistent data
void shift_atExit(void);

// code responds to #define FMM_INLINECOMPLEX (see fmm.c)

/* Note: All integer arguments in calls to the BLAS functions will be
   casted according to (int *)&BLASINT. This is to provide for the
   (bad!) platforms where sizeof(int) < sizeof(size_t) and the binary
   library uses size_t while the declaration is erroneously int. For
   such cases you will have to manually define BLASINT as size_t when
   calling the compiler. */
#ifndef BLASINT
#define BLASINT int
#endif

// BLAS-kernels
#ifdef PCWIN
#define dgemm_ dgemm
#define dtrmm_ dtrmm
#endif

// C := alpha*op( A )*op( B ) + beta*C
void dgemm_(const char *transA,const char *transB,
	    const int *M,const int *N,const int *K,
	    const double *alpha,const double *A,const int *ldA,
	    const double *B,const int *ldB,
	    const double *beta,
	    double *C,const int *ldC);

// B := alpha*op( A )*B or B := alpha*B*op( A )
void dtrmm_(const char *side,const char *uplo,
	    const char *transA,const char *diag,
	    const int *M,const int *N,
	    const double *alpha,const double *A,const int *ldA,
	    double *B,const int *ldB);

// precomputed matrices and workspace arrays
static double *binomial1,*binomial2,*binomial3;
static double *wksp1;
#ifdef FMM_INLINECOMPLEX
static double *wksp2;
#else
static dcmplx *wksp2;
#endif
static dcmplx *wksp,*rpow;

// scalars defining shift-type and sizes
static int pshift,potshift,maxm2pshift;
static double rminshift;

/*------------------------------------------------------------------------*/
void shift_alloc(int p,int pot,int maxm2p)
/* Allocates and initializes static data given a polynomial order p,
   type of potential pot and maximum number of cluster to cluster
   interactions maxm2p. */
{
  if (pshift != p || maxm2pshift < maxm2p) {
    // matrices of binomial coefficients
    mxFree(binomial1);
    binomial1 = mxCalloc(p*p,sizeof(double));
    mexMakeMemoryPersistent(binomial1);
    mxFree(binomial2);
    binomial2 = mxCalloc(p*(p+1),sizeof(double));
    mexMakeMemoryPersistent(binomial2);
    mxFree(binomial3);
    binomial3 = mxCalloc((p+1)*(p+1),sizeof(double));
    mexMakeMemoryPersistent(binomial3);

    // workspace arrays
    mxFree(wksp); // used by m2m and p2p...
    wksp = mxMalloc(4*(p+1)*sizeof(dcmplx));
    mexMakeMemoryPersistent(wksp);
    if (maxm2p < 4) maxm2p = 4;
    mxFree(wksp1); // ...m2p...
    wksp1 = mxMalloc(2*maxm2p*p*sizeof(dcmplx));
    mexMakeMemoryPersistent(wksp1);
    mxFree(wksp2); // ...m2p
    wksp2 = mxMalloc(2*maxm2p*(p+1)*sizeof(dcmplx));
    mexMakeMemoryPersistent(wksp2);

    // powers of distances
    mxFree(rpow);
    rpow = mxMalloc(maxm2p*(p+1)*sizeof(dcmplx));
    for (int i = 0; i < maxm2p; i++)
      rpow[i*(p+1)] = 1.0;
    mexMakeMemoryPersistent(rpow);

    pshift = p;
    potshift = pot;
    maxm2pshift = maxm2p;
    rminshift = exp2(-1000.0/(pshift+1));
    /* On most machines, log2(DBL_MIN) is -1022, solving for
       rmin^(pshift+1) = DBL_MIN and adding some safety yields the
       above expression (hence the constant -1000.0 above =
       DBL_MIN_EXP+21, with 21 the safety). In magnitude larger
       distances than rminshift will not produce a division by zero in
       the code. See further comment in shift_m2m() below. */
    mexAtExit(shift_atExit);

    // multipole to multipole shift
    for (int l = 0; l < p; l++) {
      binomial1[p*l] = 1.0;
      for (int k = 1; k <= l; k++)
        binomial1[p*l+k] = binomial1[p*(l-1)+k]+binomial1[p*(l-1)+k-1];
    }
    /*
    for (int k = 0; k < p; k++) {
      for (int l = 0; l < p; l++)
        mexPrintf("%f ",binomial1[p*l+k]);
      mexPrintf(" ;\n");
    }
    */

    // multipole to polynomial shift
    for (int l = 0; l <= p; l++)
      binomial2[p*l] = -1.0;
    for (int k = 1; k < p; k++) {
      binomial2[k] = -binomial2[k-1];
      for (int l = 1; l <= p; l++)
        binomial2[p*l+k] = binomial2[p*(l-1)+k]-binomial2[p*l+k-1];
    }
    /*
    for (int k = 0; k < p; k++) {
      for (int l = 0; l <= p; l++)
        mexPrintf("%f ",binomial2[p*l+k]);
      mexPrintf(" ;\n");
    }
    */

    // polynomial to polynomial shift
    binomial3[0] = 1.0;
    for (int k = 1; k <= p; k++) {
      binomial3[k] = -binomial3[k-1];
      for (int l = 1; l <= k; l++)
        binomial3[(p+1)*l+k] = binomial3[(p+1)*(l-1)+k-1]-
        binomial3[(p+1)*l+k-1];
    }
    /*
    for (int k = 0; k <= p; k++) {
      for (int l = 0; l <= p; l++)
        mexPrintf("%f ",binomial3[(p+1)*l+k]);
      mexPrintf(" ;\n");
    }
    */
  }
}
/*------------------------------------------------------------------------*/
void shift_atExit(void)
/* Deallocates persistent data. */
{
  mxFree(rpow);
  mxFree(wksp2);
  mxFree(wksp1);
  mxFree(wksp);
  mxFree(binomial3);
  mxFree(binomial2);
  mxFree(binomial1);
  pshift = 0;
}
/*------------------------------------------------------------------------*/
// snap to exactly zero
#define SNAP2ZERO(z) (fabs(creal(z)) > rminshift || \
		      fabs(cimag(z)) > rminshift ? (z) : 0.0)

void shift_m2m(dcmplx r0,dcmplx r1,dcmplx r2,dcmplx r3,
	       dcmplx *this,const dcmplx *that)
/* Multipole to multipole shift from 4 children to 1 parent; this +=
   shift(that[0...p]) {child 0} + shift(that[p+1...2*p]) {child 1} +
   ...  and so on. */
{
  // prescale
  rpow[1] = SNAP2ZERO(r0);
  rpow[pshift+2] = SNAP2ZERO(r1);
  rpow[2*(pshift+1)+1] = SNAP2ZERO(r2);
  rpow[3*(pshift+1)+1] = SNAP2ZERO(r3);

  for (int i = 0,j = 1; i < 4; i++,j++) {
    const dcmplx r = rpow[i*(pshift+1)+1];
    if (r != 0.0)
      for (int k = 0; k < pshift; k++,j++) {
	rpow[j] = rpow[j-1]*r;
	wksp[4*k+i] = that[j]/rpow[j];
      }
    else
      /* Very rare special case: child i and the parent fill up
	 exactly the same box. Zero out rpow and wksp and add the
	 contribution directly here. For potshift = 1, the zeroth
	 (log-) term is taken care of in the postscaling. */
      for (int k = 0; k < pshift; k++,j++) {
	rpow[j] = wksp[4*k+i] = 0.0;
	this[k+1] += that[j];
      }
  }

  // multiplication with lower triangular Pascal matrix
  const double one = 1.0;
  const char *si = "R",*ul = "U",*tr = "N",*di = "U";
  const BLASINT eight = 8,K = pshift;
  dtrmm_(si,ul,tr,di,(int *)&eight,(int *)&K,
	 &one,binomial1,(int *)&K,
	 (double *)wksp,(int *)&eight);

  // postscale
  if (potshift == 0)
    for (int i = 0,j = 0; i < 4; i++) {
      const dcmplx that0 = that[j];
      this[0] += that0;
      j++;
      for (int k = 0; k < pshift; k++,j++)
	this[k+1] += (wksp[4*k+i]-that0/(k+1))*rpow[j];
    }
  else
    for (int i = 0,j = 1; i < 4; i++,j++)
      for (int k = 0; k < pshift; k++,j++)
	this[k+1] += wksp[4*k+i]*rpow[j];
}
/*------------------------------------------------------------------------*/
void shift_m2ps(dcmplx z0,const void *zi,size_t siz,
		const int *ix,int begin,int end,
		const dcmplx *this1,dcmplx *this2,
		const dcmplx *that1,dcmplx *that2)
/* Symmetric multipole to polynomial shift; this2 +=
   shift(that1[ix[i]]) followed by that2[ix[i]] += shift(this1) for
   index i in [begin,end). Center coordinates of the boxes are given
   in z0 and zi (in which each entry is separated by siz bytes).

   A somewhat singular feature is that a negative index in ix,
   pointing to a box at a previous level, is detected and implies that
   the 3 following indices are ignored. This ensures that connections
   between levels are properly handled and that the connectivity
   indices are convenient to assemble. */
{
  int n = end-begin; // not const; might change below
  if (n <= 0) return; // dgemm_() below fails when M is 0

#ifdef FMM_INLINECOMPLEX
  // prescale
  for (int i = begin,ii = 0; i < end; i++,ii++) {
    const dcmplx r0 = 1.0/(*(dcmplx *)(zi+ix[i]*siz)-z0);
    const double re = creal(r0),im = cimag(r0);
    double *ri = (double *)&rpow[ii*(pshift+1)];
    const dcmplx *that1i = &that1[ix[i]*(pshift+1)];
    double pre,pim; // register variables for complex multiplication

    for (int j = 1; j < pshift; ) {
      ri[2*j] = ri[2*(j-1)]*re-ri[2*(j-1)+1]*im;
      ri[2*j+1] = ri[2*(j-1)]*im+ri[2*(j-1)+1]*re;
      pre = creal(that1i[j])*ri[2*j]-cimag(that1i[j])*ri[2*j+1];
      pim = creal(that1i[j])*ri[2*j+1]+cimag(that1i[j])*ri[2*j];
      wksp1[pshift*4*ii+j-1] = pre;
      wksp1[pshift*(4*ii+1)+j-1] = pim;
      pre = -creal(this1[j])*ri[2*j]+cimag(this1[j])*ri[2*j+1];
      pim = -creal(this1[j])*ri[2*j+1]-cimag(this1[j])*ri[2*j];
      wksp1[pshift*(4*ii+2)+j-1] = pre;
      wksp1[pshift*(4*ii+3)+j-1] = pim;
      j++;

      ri[2*j] = ri[2*(j-1)]*re-ri[2*(j-1)+1]*im;
      ri[2*j+1] = ri[2*(j-1)]*im+ri[2*(j-1)+1]*re;
      pre = creal(that1i[j])*ri[2*j]-cimag(that1i[j])*ri[2*j+1];
      pim = creal(that1i[j])*ri[2*j+1]+cimag(that1i[j])*ri[2*j];
      wksp1[pshift*4*ii+j-1] = pre;
      wksp1[pshift*(4*ii+1)+j-1] = pim;
      pre = creal(this1[j])*ri[2*j]-cimag(this1[j])*ri[2*j+1];
      pim = creal(this1[j])*ri[2*j+1]+cimag(this1[j])*ri[2*j];
      wksp1[pshift*(4*ii+2)+j-1] = pre;
      wksp1[pshift*(4*ii+3)+j-1] = pim;
      j++;
    }
    if (pshift&1) {
      ri[2*pshift] = ri[2*(pshift-1)]*re-ri[2*(pshift-1)+1]*im;
      ri[2*pshift+1] = ri[2*(pshift-1)]*im+ri[2*(pshift-1)+1]*re;
      pre = creal(that1i[pshift])*ri[2*pshift]-cimag(that1i[pshift])*ri[2*pshift+1];
      pim = creal(that1i[pshift])*ri[2*pshift+1]+cimag(that1i[pshift])*ri[2*pshift];
      wksp1[pshift*4*ii+pshift-1] = pre;
      wksp1[pshift*(4*ii+1)+pshift-1] = pim;
      pre = -creal(this1[pshift])*ri[2*pshift]+cimag(this1[pshift])*ri[2*pshift+1];
      pim = -creal(this1[pshift])*ri[2*pshift+1]-cimag(this1[pshift])*ri[2*pshift];
      wksp1[pshift*(4*ii+2)+pshift-1] = pre;
      wksp1[pshift*(4*ii+3)+pshift-1] = pim;
    }
    if (ix[i] < 0) {
      i += 3;
      n -= 3;
    }
  }
#else // ~FMM_INLINECOMPLEX
  // prescale
  for (int i = begin,ii = 0; i < end; i++,ii++) {
    const dcmplx r0 = 1.0/(*(dcmplx *)(zi+ix[i]*siz)-z0);
    dcmplx *ri = &rpow[ii*(pshift+1)];
    const dcmplx *that1i = &that1[ix[i]*(pshift+1)];
    dcmplx mul;

    for (int j = 1; j < pshift; ) {
      ri[j] = ri[j-1]*r0;
      mul = that1i[j]*ri[j];
      wksp1[pshift*4*ii+j-1] = creal(mul);
      wksp1[pshift*(4*ii+1)+j-1] = cimag(mul);
      mul = -this1[j]*ri[j];
      wksp1[pshift*(4*ii+2)+j-1] = creal(mul);
      wksp1[pshift*(4*ii+3)+j-1] = cimag(mul);
      j++;
      ri[j] = ri[j-1]*r0;
      mul = that1i[j]*ri[j];
      wksp1[pshift*4*ii+j-1] = creal(mul);
      wksp1[pshift*(4*ii+1)+j-1] = cimag(mul);
      mul = this1[j]*ri[j];
      wksp1[pshift*(4*ii+2)+j-1] = creal(mul);
      wksp1[pshift*(4*ii+3)+j-1] = cimag(mul);
      j++;
    }
    if (pshift&1) {
      ri[pshift] = ri[pshift-1]*r0;
      mul = that1i[pshift]*ri[pshift];
      wksp1[pshift*4*ii+pshift-1] = creal(mul);
      wksp1[pshift*(4*ii+1)+pshift-1] = cimag(mul);
      mul = -this1[pshift]*ri[pshift];
      wksp1[pshift*(4*ii+2)+pshift-1] = creal(mul);
      wksp1[pshift*(4*ii+3)+pshift-1] = cimag(mul);
    }
    if (ix[i] < 0) {
      i += 3;
      n -= 3;
    }
  }
#endif // FMM_INLINECOMPLEX

  // multiplication with rectangular Pascal matrix
#ifndef FMM_INLINECOMPLEX
  const double one = 1.0,zero = 0.0;
  const char *tr1 = "T",*tr2 = "N";
  const BLASINT M = 4*n,N = pshift+1,K = pshift;
  dgemm_(tr1,tr2,(int *)&M,(int *)&N,(int *)&K,
	 &one,wksp1,(int *)&K,
	 binomial2,(int *)&K,
	 &zero,(double *)wksp2,(int *)&M);
#else // ~FMM_INLINECOMPLEX
  const double one = 1.0,zero = 0.0;
  const char *tr1 = "T",*tr2 = "N";
  const BLASINT M = pshift+1,N = 4*n,K = pshift;
  dgemm_(tr1,tr2,(int *)&M,(int *)&N,(int *)&K,
	 &one,binomial2,(int *)&K,
	 wksp1,(int *)&K,
	 &zero,wksp2,(int *)&M);
#endif // FMM_INLINECOMEPLEX

  // postscale
#ifdef FMM_INLINECOMPLEX
  if (potshift == 0)
    for (int i = begin,ii = 0; i < end; i++,ii++) {
      const int pshift1 = pshift+1;
      double *ri = (double *)&rpow[ii*pshift1];
      dcmplx *that2i = &that2[ix[i]*pshift1];
      const double rec0 = creal(this1[0]),imc0 = cimag(this1[0]);
      const double red0 = creal(that1[ix[i]*pshift1]),
	imd0 = cimag(that1[ix[i]*pshift1]);
      double pre,pim; // register variables for complex multiplication

      {
	const dcmplx logr = clog(ri[2]+I*ri[3]);
	this2[0] += wksp2[pshift1*4*ii]+I*wksp2[pshift1*(4*ii+1)]
	  -(red0*creal(logr)-imd0*cimag(logr)+
	    I*(imd0*creal(logr)+red0*cimag(logr)));
	that2i[0] += wksp2[pshift1*(4*ii+2)]+I*wksp2[pshift1*(4*ii+3)]
	  -(rec0*creal(logr)-imc0*(cimag(logr)-M_PI)+
	    I*(imc0*creal(logr)+rec0*(cimag(logr)-M_PI)));
      }
      for (int j = 1; j < pshift; ) {
	pre = (wksp2[pshift1*4*ii+j]-red0/j)*ri[2*j]-
	  (wksp2[pshift1*(4*ii+1)+j]-imd0/j)*ri[2*j+1];
	pim = (wksp2[pshift1*4*ii+j]-red0/j)*ri[2*j+1]+
	  (wksp2[pshift1*(4*ii+1)+j]-imd0/j)*ri[2*j];
	this2[j] += pre+I*pim;
	pre = (wksp2[pshift1*(4*ii+2)+j]-rec0/j)*ri[2*j]-
	  (wksp2[pshift1*(4*ii+3)+j]-imc0/j)*ri[2*j+1];
	pim = (wksp2[pshift1*(4*ii+2)+j]-rec0/j)*ri[2*j+1]+
	  (wksp2[pshift1*(4*ii+3)+j]-imc0/j)*ri[2*j];
	that2i[j] -= pre+I*pim;
	j++;

	pre = (wksp2[pshift1*4*ii+j]-red0/j)*ri[2*j]-
	  (wksp2[pshift1*(4*ii+1)+j]-imd0/j)*ri[2*j+1];
	pim = (wksp2[pshift1*4*ii+j]-red0/j)*ri[2*j+1]+
	  (wksp2[pshift1*(4*ii+1)+j]-imd0/j)*ri[2*j];
	this2[j] += pre+I*pim;
	pre = (wksp2[pshift1*(4*ii+2)+j]-rec0/j)*ri[2*j]-
	  (wksp2[pshift1*(4*ii+3)+j]-imc0/j)*ri[2*j+1];
	pim = (wksp2[pshift1*(4*ii+2)+j]-rec0/j)*ri[2*j+1]+
	  (wksp2[pshift1*(4*ii+3)+j]-imc0/j)*ri[2*j];
	that2i[j] += pre+I*pim;
	j++;
      }
      if (pshift&1) {
	pre = (wksp2[pshift1*4*ii+pshift]-red0/pshift)*ri[2*pshift]-
	  (wksp2[pshift1*(4*ii+1)+pshift]-imd0/pshift)*ri[2*pshift+1];
	pim = (wksp2[pshift1*4*ii+pshift]-red0/pshift)*ri[2*pshift+1]+
	  (wksp2[pshift1*(4*ii+1)+pshift]-imd0/pshift)*ri[2*pshift];
	this2[pshift] += pre+I*pim;
	pre = (wksp2[pshift1*(4*ii+2)+pshift]-rec0/pshift)*ri[2*pshift]-
	  (wksp2[pshift1*(4*ii+3)+pshift]-imc0/pshift)*ri[2*pshift+1];
	pim = (wksp2[pshift1*(4*ii+2)+pshift]-rec0/pshift)*ri[2*pshift+1]+
	  (wksp2[pshift1*(4*ii+3)+pshift]-imc0/pshift)*ri[2*pshift];
	that2i[pshift] -= pre+I*pim;
      }
      if (ix[i] < 0) i += 3;
    }
  else 
    for (int i = begin,ii = 0; i < end; i++,ii++) {
      const int pshift1 = pshift+1;
      double *ri = (double *)&rpow[ii*pshift1];
      dcmplx *that2i = &that2[ix[i]*pshift1];
      double pre,pim;

      for (int j = 0; j < pshift; ) {
	pre = wksp2[pshift1*4*ii+j]*ri[2*j]-wksp2[pshift1*(4*ii+1)+j]*ri[2*j+1];
	pim = wksp2[pshift1*4*ii+j]*ri[2*j+1]+wksp2[pshift1*(4*ii+1)+j]*ri[2*j];
	this2[j] += pre+I*pim;
	pre = wksp2[pshift1*(4*ii+2)+j]*ri[2*j]-wksp2[pshift1*(4*ii+3)+j]*ri[2*j+1];
	pim = wksp2[pshift1*(4*ii+2)+j]*ri[2*j+1]+wksp2[pshift1*(4*ii+3)+j]*ri[2*j];
	that2i[j] += pre+I*pim;
	j++;

	pre = wksp2[pshift1*4*ii+j]*ri[2*j]-wksp2[pshift1*(4*ii+1)+j]*ri[2*j+1];
	pim = wksp2[pshift1*4*ii+j]*ri[2*j+1]+wksp2[pshift1*(4*ii+1)+j]*ri[2*j];
	this2[j] += pre+I*pim;
	pre = wksp2[pshift1*(4*ii+2)+j]*ri[2*j]-wksp2[pshift1*(4*ii+3)+j]*ri[2*j+1];
	pim = wksp2[pshift1*(4*ii+2)+j]*ri[2*j+1]+wksp2[pshift1*(4*ii+3)+j]*ri[2*j];
	that2i[j] -= pre+I*pim;
	j++;
      }
      if (~pshift&1) {
	pre = wksp2[pshift1*4*ii+pshift]*ri[2*pshift]-
	  wksp2[pshift1*(4*ii+1)+pshift]*ri[2*pshift+1];
	pim = wksp2[pshift1*4*ii+pshift]*ri[2*pshift+1]+
	  wksp2[pshift1*(4*ii+1)+pshift]*ri[2*pshift];
	this2[pshift] += pre+I*pim;
	pre = wksp2[pshift1*(4*ii+2)+pshift]*ri[2*pshift]-
	  wksp2[pshift1*(4*ii+3)+pshift]*ri[2*pshift+1];
	pim = wksp2[pshift1*(4*ii+2)+pshift]*ri[2*pshift+1]+
	  wksp2[pshift1*(4*ii+3)+pshift]*ri[2*pshift];
	that2i[pshift] += pre+I*pim;
      }
      if (ix[i] < 0) i += 3;
    }
#else // ~FMM_INLINECOMPLEX
  if (potshift == 0)
    for (int i = begin,ii = 0; i < end; i++,ii++) {
      dcmplx *ri = &rpow[ii*(pshift+1)];
      dcmplx *that2i = &that2[ix[i]*(pshift+1)];
      const dcmplx this0 = this1[0];
      const dcmplx that0 = that1[ix[i]*(pshift+1)];

      this2[0] += wksp2[2*ii]-that0*clog(ri[1]);
      that2i[0] += wksp2[2*ii+1]-this0*clog(-ri[1]); // *** can avoid the 2nd call
      for (int j = 1; j < pshift; ) {
	this2[j] += (wksp2[2*(n*j+ii)]-that0/j)*ri[j];
	that2i[j] -= (wksp2[2*(n*j+ii)+1]-this0/j)*ri[j];
	j++;
	this2[j] += (wksp2[2*(n*j+ii)]-that0/j)*ri[j];
	that2i[j] += (wksp2[2*(n*j+ii)+1]-this0/j)*ri[j];
	j++;
      }
      if (pshift&1) {
	this2[pshift] += (wksp2[2*(n*pshift+ii)]-that0/pshift)*ri[pshift];
	that2i[pshift] -= (wksp2[2*(n*pshift+ii)+1]-this0/pshift)*ri[pshift];
      }
      if (ix[i] < 0) i += 3;
    }
  else 
    for (int i = begin,ii = 0; i < end; i++,ii++) {
      dcmplx *ri = &rpow[ii*(pshift+1)];
      dcmplx *that2i = &that2[ix[i]*(pshift+1)];

      for (int j = 0; j < pshift; ) {
	this2[j] += wksp2[2*(n*j+ii)]*ri[j];
	that2i[j] += wksp2[2*(n*j+ii)+1]*ri[j];
	j++;
	this2[j] += wksp2[2*(n*j+ii)]*ri[j];
	that2i[j] -= wksp2[2*(n*j+ii)+1]*ri[j];
	j++;
      }
      if (~pshift&1) {
	this2[pshift] += wksp2[2*(n*pshift+ii)]*ri[pshift];
	that2i[pshift] += wksp2[2*(n*pshift+ii)+1]*ri[pshift];
      }
      if (ix[i] < 0) i += 3;
    }
#endif // FMM_INLINECOMPLEX
}
/*------------------------------------------------------------------------*/
void shift_p2p(dcmplx r0,dcmplx r1,dcmplx r2,dcmplx r3,
	       dcmplx *this,const dcmplx *that)
/* Polynomial to polynomial shift from 1 parent to 4 children;
   this[0..p] += shift(that) {child 0}, this[p+1..2*p] += shift(that)
   {child 1}, and so on. */
{
  // prescale
  rpow[1] = SNAP2ZERO(r0);
  rpow[pshift+2] = SNAP2ZERO(r1);
  rpow[2*(pshift+1)+1] = SNAP2ZERO(r2);
  rpow[3*(pshift+1)+1] = SNAP2ZERO(r3);
  for (int i = 0,j = 0; i < 4; i++) {
    const dcmplx r = rpow[i*(pshift+1)+1];
    for (int k = 0; k < pshift; k++,j++) {
      wksp[4*k+i] = that[k]*rpow[j];
      rpow[j+1] = rpow[j]*r;
    }
    wksp[4*pshift+i] = that[pshift]*rpow[j];
    j++;
  }

  // multiplication with upper triangular Pascal matrix
  const double one = 1.0;
  const char *si = "R",*ul = "L",*tr = "N",*di = "U";
  const BLASINT eight = 8,N = pshift+1;
  dtrmm_(si,ul,tr,di,(int *)&eight,(int *)&N,
	 &one,binomial3,(int *)&N,
	 (double *)wksp,(int *)&eight);

  // postscale
  for (int i = 0,j = 0; i < 4; i++) {
    if (rpow[j+1] != 0.0)
      for (int k = 0; k <= pshift; k++,j++)
	this[j] += wksp[4*k+i]/rpow[j];
    else // again, the rare special case
      for (int k = 0; k <= pshift; k++,j++)
	this[j] += that[k];
  }
}
/*------------------------------------------------------------------------*/
