function h = meshplot(M,varargin)
%MESHPLOT Plot FMM mesh.
%   MESHPLOT(M) plots the FMM mesh M as obtained via the 'out'
%   property in FMM2D.
%
%   H = MESPLOT(M,...) sends additional arguments to PLOT and returns
%   a handle to the resulting plot.
%
%   Example:
%     N = 1000;
%     z = complex(rand(N,1),rand(N,1));
%     m = rand(N,1);
%
%     [M,C] = fmm2d(z,m,'tol',1e-6,'out','mesh');
%     % "symmetrize" C:
%     for i = 1:length(C), C{i} = speye(size(C{i}))+C{i}''+C{i}'; end
%     % (note: the double transpose is needed on some platforms
%     %  since the C's aren't sorted)
%     meshplot(M{end},'b:'); % finest level
%     hold on, plot(z,'b.'); axis square
%
%     % plot connections of box i at level l
%     figure, hold on,
%     l = 5;
%     i = 100;
%     i1 = find(C{l}(:,i) == 1);
%     i2 = find(C{l}(:,i) == 2);
%     j1 = find(C{l}(:,i) == -1);
%     j2 = find(C{l}(:,i) == -2);
%     j2a = floor((j2(j2 > i)-1)/4)+1;
%     j2b = j2(j2 < i);
%     meshplot(M{l-1}(:,j2a),'g-.'); % box i to boxes j2a (at previous level)
%     meshplot(M{l}(:,j2b),'g--');   % parent of i to boxes j2b
%     meshplot(M{l}(:,j1),'b');      % box i to boxes j1 (this level)
%     meshplot(M{l}(:,i2),'c-');     % quasi cluster interaction
%     meshplot(M{l}(:,i1),'r-');     % direct interaction
%     meshplot(M{l}(:,i),'k*-');     % box i itself
%     axis equal
%
%   See also FMM2D, MESHSTATS, MESHQUAL, PLOT.

% S. Engblom 2009-10-01 (Minor revision)
% S. Engblom 2009-05-08

z = M([1 1 1 1],:)+[M(2,:); conj(M(2,:)); -M(2,:); -conj(M(2,:))];
if nargout > 0
  h = plot(z([1:end 1],:),varargin{:});
else
  plot(z([1:end 1],:),varargin{:});
end
